import gc
import os
from dataclasses import dataclass
from typing import List, Optional

import torch
from external_models import *
from fastapi import FastAPI, HTTPException
from leanprogress import LeanProgressScorer
from models import *
from pydantic import BaseModel

app = FastAPI()


@dataclass
class ModelBundle:
    generator: Generator
    progress_scorer: Optional[LeanProgressScorer] = None


def _build_progress_scorer() -> Optional[LeanProgressScorer]:
    model_path = os.getenv("LEANPROGRESS_MODEL", "").strip()
    if not model_path:
        return None
    device = os.getenv("LEANPROGRESS_DEVICE", "auto").strip()
    template = os.getenv("LEANPROGRESS_TEMPLATE", None)
    try:
        return LeanProgressScorer(
            model_name=model_path,
            device=device,
            template=template,
        )
    except Exception as exc:
        raise RuntimeError(
            f"Failed to load LeanProgress model '{model_path}': {exc}"
        ) from exc


models = {
    "deepseek": ModelBundle(
        generator=HFTacticGenerator(
            model="deepseek-ai/DeepSeek-Prover-V2-671B:novita",
            temperature=0.6,
            max_new_tokens=256,
            top_p=0.9,
            length_penalty=0,
            num_return_sequences=4,
            do_sample=True,
            output_scores=True,
            output_logits=False,
            return_dict_in_generate=True,
            device="auto",
        ),
        progress_scorer=_build_progress_scorer(),
    ),
}


class GeneratorRequest(BaseModel):
    name: str
    input: str
    prefix: Optional[str]
    use_reward: bool = False


class Generation(BaseModel):
    output: str
    score: float
    model_score: Optional[float] = None
    steps_remaining: Optional[float] = None
    reward: Optional[float] = None


class GeneratorResponse(BaseModel):
    outputs: List[Generation]


class EncoderRequest(BaseModel):
    name: str
    input: str


class EncoderResponse(BaseModel):
    outputs: List[float]


@app.post("/generate")
async def generate(req: GeneratorRequest) -> GeneratorResponse:
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        gc.collect()

    try:
        bundle = models[req.name]
        target_prefix = req.prefix if req.prefix is not None else ""
        outputs = bundle.generator.generate(req.input, target_prefix)
        generations: List[Generation] = []
        if req.use_reward:
            if bundle.progress_scorer is None:
                raise HTTPException(
                    status_code=400,
                    detail="LeanProgress scoring requested but no model configured.",
                )
            for out, model_score in outputs:
                steps_remaining = bundle.progress_scorer.predict(
                    goal=req.input,
                    tactic=out,
                    prefix=req.prefix,
                )
                reward_value = -steps_remaining
                generations.append(
                    Generation(
                        output=out,
                        score=reward_value,
                        model_score=model_score,
                        steps_remaining=steps_remaining,
                        reward=reward_value,
                    )
                )
        else:
            generations = [
                Generation(output=out, score=score) for out, score in outputs
            ]

        if torch.cuda.is_available():
            torch.cuda.empty_cache()
            gc.collect()
        return GeneratorResponse(outputs=generations)
    except torch.cuda.OutOfMemoryError:
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
            gc.collect()
        raise HTTPException(status_code=500, detail="GPU out of memory")
    except Exception as e:
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
            gc.collect()
        raise HTTPException(status_code=500, detail=str(e))


@app.post("/encode")
async def encode(req: EncoderRequest) -> EncoderResponse:
    bundle = models[req.name]
    if not hasattr(bundle.generator, "encode"):
        raise HTTPException(
            status_code=400,
            detail=f"Model '{req.name}' does not support encoding requests.",
        )
    feature = bundle.generator.encode(req.input)  # type: ignore[attr-defined]
    return EncoderResponse(outputs=feature.tolist())



"""This module defines traced repos/files/theorems."""

import itertools
import json
import os
import random
import re
import webbrowser
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import networkx as nx
from loguru import logger
from lxml import etree
from tqdm import tqdm

from lean_dojo_v2.utils.constants import (
    LEAN4_PACKAGES_DIR,
    LOAD_USED_PACKAGES_ONLY,
    NUM_WORKERS,
)
from lean_dojo_v2.utils.filesystem import compute_md5, is_git_repo
from lean_dojo_v2.utils.lean import to_dep_path, to_json_path, to_lean_path, to_xml_path

from .ast import *
from .lean import LeanFile, LeanGitRepo, Pos, Theorem


@dataclass(frozen=True)
class Comment:
    """A comment in a Lean file."""

    start: Pos
    end: Pos
    text: str

    def __post_init__(self) -> None:
        assert isinstance(self.start, Pos)
        assert isinstance(self.end, Pos)
        assert self.start <= self.end
        assert isinstance(self.text, str)

    def to_xml(self, parent: etree.Element) -> None:
        tree = etree.SubElement(parent, self.__class__.__name__)
        tree.set("start", str(self.start))
        tree.set("end", str(self.end))
        tree.set("text", self.text)

    @classmethod
    def from_xml(cls, tree: etree.Element) -> "Comment":
        start = Pos.from_str(tree.attrib["start"])
        end = Pos.from_str(tree.attrib["end"])
        text = tree.attrib["text"]
        return cls(start, end, text)


def _collect_lean4_comments(ast: FileNode) -> List[Comment]:
    comments = []

    def _callback(node, _):
        nonlocal comments
        if isinstance(node, CommandModuledocNode) or isinstance(
            node, CommandDoccommentNode
        ):
            comments.append(Comment(node.start, node.end, node.comment))
        elif is_leaf(node) and node.trailing.strip().startswith("--"):
            num_spaces = node.trailing.index("--")
            text = node.trailing[num_spaces:]
            start = node.lean_file.offset(node.end, num_spaces)
            end = node.lean_file.offset(start, len(text))
            comments.append(Comment(start, end, text))

    ast.traverse_preorder(_callback, node_cls=None)
    return comments


_SINGLE_LINE_COMMENT_REGEX = r"--.*?(\n|$)"
_MULTI_LINE_COMMENT_REGEX = r"/-.*?(-/|$)"
_COMMENT_REGEX = re.compile(
    f"{_SINGLE_LINE_COMMENT_REGEX}|{_MULTI_LINE_COMMENT_REGEX}", re.DOTALL
)


def get_code_without_comments(
    lean_file: LeanFile, start: Pos, end: Pos, comments: List[Comment]
) -> str:
    """Return the code in ``lean_file`` from ``start`` to ``end`` with comments removed.

    Args:
        lean_file (LeanFile): The lean source file.
        start (Pos): The start position.
        end (Pos): The end position.
        comments (List[Comment]): A list of :class:`Comment` objects.

    Returns:
        str: Human-written code with comments removed.
    """
    base = start
    code_segs = []

    for c in comments:
        if base <= c.start and c.end <= end:
            code_segs.append(lean_file[base : c.start])
            base = c.end

    code_segs.append(lean_file[base:end])
    code = "".join(code_segs)

    code = _COMMENT_REGEX.sub("", code)
    assert "--" not in code and "/-" not in code

    return code.strip()


@dataclass(frozen=True)
class TracedTactic:
    """A traced tactic is a tactic annotated with additional information including
    its AST and the states before/after the tactic.
    """

    ast: Node = field(repr=False)
    """AST of the tactic.
    """

    traced_theorem: Optional["TracedTheorem"] = field(
        default=None, repr=False, compare=False
    )
    """The traced theorem this tactic belongs to.
    """

    def __getstate__(self) -> Dict[str, Any]:
        d = {k: v for k, v in self.__dict__.items() if k != "traced_theorem"}
        d["traced_theorem"] = None  # Avoid serializing the traced theorem.
        return d

    @property
    def tactic(self) -> str:
        """The raw tactic string."""
        return self.ast.tactic

    @property
    def state_before(self) -> str:
        """Pretty-printed state before applying the tactic."""
        assert self.ast.state_before is not None
        return self.ast.state_before

    @property
    def state_after(self) -> str:
        """Pretty-printed state after applying the tactic."""
        assert self.ast.state_after is not None
        return self.ast.state_after

    @property
    def start(self) -> Pos:
        """Start position in :file:`*.lean` file."""
        return self.ast.start

    @property
    def end(self) -> Pos:
        """End position in :file:`*.lean` file."""
        return self.ast.end

    def to_string(self) -> str:
        return f"{self.__class__.__name__}(tactic={self.tactic}, state_before={self.state_before}, state_after={self.state_after})"

    def __str__(self) -> str:
        return self.to_string()

    def __repr__(self) -> str:
        return self.to_string()

    def get_annotated_tactic(self) -> Tuple[str, List[Dict[str, Any]]]:
        """Return the tactic annotated with premise information.

        Premises in the tactic are marked by ``<a> ... </a>``. For example,
        :code:`rw [add_comm b]` contains a premise :code:`add_comm` and therefore
        becomes :code:`rw [<a>add_comm</a> b]`. In addition, the function returns
        the provenance (full name, file path, line/column numbers) of all premises.

        Returns:
            Tuple[str, List[Dict[str, Any]]]: The first return value is the tactic string marked by ``<a> ... </a>``. The second return value is a list of provenances.
        """
        assert self.traced_theorem != None
        lean_file = self.traced_theorem.traced_file.lean_file
        annot_tac = []
        provenances = []
        cur = self.start

        def _callback4(node: IdentNode, _):
            nonlocal cur

            if (
                node.full_name is not None
                and node.mod_name is not None
                and node.def_start is not None
                and node.def_end is not None
            ):
                if cur <= node.start:
                    annot_tac.append(lean_file[cur : node.start])
                    annot_tac.append("<a>" + lean_file[node.start : node.end] + "</a>")
                    prov = {"full_name": node.full_name}
                    prov["def_path"] = node.def_path
                    prov["def_pos"] = list(node.def_start)
                    prov["def_end_pos"] = list(node.def_end)
                    provenances.append(prov)
                    cur = node.end

        self.ast.traverse_preorder(_callback4, IdentNode)
        annot_tac.append(lean_file[cur : self.end])

        return "".join(annot_tac), provenances


@dataclass(frozen=True)
class TracedTheorem:
    """A traced theorem is a theorem with additional information such as the AST."""

    root_dir: Path = field(repr=False)
    """Root directory of the corresponding traced repo.
    """

    theorem: Theorem
    """The corresponding :class:`Theorem` object.
    """

    ast: Union[CommandTheoremNode, LemmaNode, MathlibTacticLemmaNode] = field(
        repr=False, compare=False
    )
    """AST of the theorem.
    """

    comments: List[Comment] = field(repr=False, compare=False)
    """All comments in the theorem/proof.
    """

    traced_file: Optional["TracedFile"] = field(default=None, repr=False, compare=False)
    """The traced file this theorem belongs to.
    """

    def __post_init__(self) -> None:
        assert (
            self.root_dir.is_absolute() and self.root_dir == self.traced_file.root_dir
        )

    def __getstate__(self) -> Dict[str, Any]:
        d = {k: v for k, v in self.__dict__.items() if k != "traced_file"}
        d["traced_file"] = None
        return d

    @property
    def start(self) -> Pos:
        """Start position in :file:`*.lean` file."""
        return self.ast.start

    @property
    def end(self) -> Pos:
        """End position in :file:`*.lean` file."""
        return self.ast.end

    @property
    def repo(self) -> LeanGitRepo:
        """The Lean repo this theorem belongs to."""
        return self.theorem.repo

    @property
    def file_path(self) -> Path:
        """The theorem's file path (relative to the root directory)."""
        return self.theorem.file_path

    @property
    def traced_repo(self) -> "TracedRepo":
        """The traced repo this theorem belongs to."""
        if self.traced_file is None:
            return None
        else:
            return self.traced_file.traced_repo

    @property
    def is_private(self) -> bool:
        """Check if the theorem is private."""
        return self.ast.is_private()

    def show(self) -> None:
        """Show the theorem in the default browser."""
        url = os.path.join(
            self.repo.url,
            "blob",
            self.repo.commit,
            self.file_path,
            f"#L{self.start.line_nb}-L{self.end.line_nb}",
        )
        webbrowser.open(url)

    def has_tactic_proof(self) -> bool:
        """Check if the theorem has a tactic-style proof."""
        return self.ast.has_tactic_proof()

    def get_proof_node(self) -> Node:
        """Return the AST of the theorem's proof."""
        return self.ast.get_proof_node()

    def locate_proof(self) -> Tuple[Pos, Pos]:
        """Return the start/end positions of the proof."""
        start, end = self.get_proof_node().get_closure()
        if end < self.end:
            end = self.end
        return start, end

    def get_tactic_proof(self) -> Optional[str]:
        """Return the tactic-style proof (if any)."""
        if not self.has_tactic_proof():
            return None
        node = self.get_proof_node()
        start, end = node.get_closure()
        proof = get_code_without_comments(node.lean_file, start, end, self.comments)
        if not re.match(r"^(by|begin)\s", proof):
            return None
        else:
            return proof

    def get_theorem_statement(self) -> str:
        """Return the theorem statement."""
        proof_start, _ = self.locate_proof()
        return get_code_without_comments(
            self.traced_file.lean_file, self.ast.start, proof_start, self.comments
        )

    def get_single_tactic_proof(self) -> Optional[str]:
        """Wrap the proof into a single (potentially very long) tactic."""
        if not self.has_tactic_proof():
            return None
        node = self.get_proof_node()
        start, end = node.get_closure()
        proof = get_code_without_comments(node.lean_file, start, end, self.comments)

        raise NotImplementedError
        assert isinstance(node.children[0], AtomNode) and node.children[0].val == "by"
        assert proof.startswith("by")
        proof = proof[len("by") :].strip()

        return proof

    def get_premise_full_names(self) -> List[str]:
        """Return the fully qualified names of all premises used in the proof."""
        names = []

        def _callback(node: IdentNode, _: List[Node]):
            if node.full_name is not None:
                names.append(node.full_name)

        self.ast.traverse_preorder(_callback, node_cls=IdentNode)

        return names

    def get_traced_tactics(self, atomic_only: bool = False) -> List[TracedTactic]:
        """Return a list of traced tactics in the proof."""
        tacs = self._get_traced_tactics_lean4(atomic_only)

        # Deduplicate.
        signatures = set()
        tacs_dedup = []
        for t in tacs:
            sig = (t.state_before, t.tactic, t.state_after)
            if sig not in signatures:
                signatures.add(sig)
                tacs_dedup.append(t)

        return tacs_dedup

    def _get_traced_tactics_lean4(
        self, atomic_only: bool = False
    ) -> List[TracedTactic]:
        tacs = []

        def _callback(node, _):
            if not isinstance(
                node,
                (
                    TacticTacticseq1IndentedNode,
                    TacticTacticseqbracketedNode,
                ),
            ):
                return
            for tac_node in node.get_tactic_nodes(atomic_only):
                if (
                    hasattr(tac_node, "state_before")
                    and tac_node.state_before is not None
                ):
                    # Tactics outside theorem/lemma definitions are not recorded.
                    tacs.append(TracedTactic(tac_node, self))

        self.ast.traverse_preorder(_callback, node_cls=None)
        return tacs

    def get_num_tactics(self) -> int:
        """Return the number of tactics in the proof."""
        return len(self.get_traced_tactics())


_TAG_INDEX_REGEX = re.compile(r"(?P<key>\S+)\[(?P<idx>\d+)\]$")


def _qualify_name(name: str, prefix: str) -> str:
    """Qualify a name with a prefix."""
    if name.startswith("_root_."):
        return name[len("_root_.") :]
    elif prefix == "":
        return name
    else:
        return f"{prefix}.{name}"


def _fix_indentation(tac: str, indent: int) -> str:
    """Fix the indentation of a tactic."""
    lines = tac.splitlines()
    if len(lines) == 1:
        return tac
    else:
        lines_new = [lines[0]]
        for l in lines[1:]:
            for i in range(len(l)):
                if l[i] != " " or i >= indent:
                    lines_new.append(l[i:])
                    break

        return "\n".join(lines_new)


@dataclass(eq=False)
class TracedFile:
    """A traced file is a Lean source file annotated with syntactic/semantic information
    such as tactic states, Lean expressions, and abstract syntax trees (ASTs).
    """

    root_dir: Path
    """Root directory (in absolute path) of the corresponding traced repo.
    """

    repo: LeanGitRepo
    """The Lean repo this traced file belongs to.
    """

    lean_file: LeanFile
    """Lean source file of this traced file.
    """

    ast: FileNode = field(repr=False)
    """Abstract syntax tree (AST) of the entire :code:`*.lean` file.
    
    AST nodes are defined in :ref:`lean_dojo.data_extraction.ast`. 
    """

    comments: List[Comment] = field(repr=False)
    """All comments in the :code:`*.lean` file.
    """

    traced_repo: Optional["TracedRepo"] = field(default=None, repr=False)
    """The traced repo this traced file belongs to.
    
    Note that ``traced_repo`` will become None after the traced file is serialized/deserialized on its own.
    """

    def __post_init__(self) -> None:
        assert self.root_dir.is_absolute(), f"{self.root_dir} is not an absolute path"

    def __getstate__(self) -> Dict[str, Any]:
        d = {k: v for k, v in self.__dict__.items() if k != "traced_repo"}
        d["traced_repo"] = None
        return d

    @property
    def path(self) -> Path:
        """Path of the :file:`*.lean` file relative to the root directory."""
        return self.lean_file.path

    @property
    def abs_path(self) -> Path:
        """Absolute path of the :code:`*.lean` file."""
        return self.root_dir / self.path

    @property
    def has_prelude(self) -> bool:
        """Check whether the file starts with :code:``prelude``.

        :code:``prelude`` instructs Lean NOT to include its built-in library automatically.
        """
        result = False

        def _callback(node: ModulePreludeNode, _: List[Node]):
            nonlocal result
            result = True
            return True  # Stop traversing.

        self.ast.traverse_preorder(_callback, node_cls=ModulePreludeNode)
        return result

    @classmethod
    def from_traced_file(
        cls, root_dir: Union[str, Path], json_path: Path, repo: LeanGitRepo
    ) -> "TracedFile":
        """Construct a :class:`TracedFile` object by parsing a :file:`*.ast.json` file
        produced by :code:`lean --ast --tsast --tspp` (Lean 3) or :file:`ExtractData.lean` (Lean 4).

        Args:
            root_dir (Union[str, Path]): Root directory of the traced repo.
            json_path (Path): Path of the :file:`*.ast.json` file relative to ``root_dir``.
        """
        root_dir = Path(root_dir)
        root_dir = root_dir.resolve()
        if not json_path.is_absolute():
            json_path = root_dir / json_path
        if not json_path.exists():
            raise FileNotFoundError(f"{json_path} does not exist")
        assert json_path.suffixes == [
            ".ast",
            ".json",
        ], f"{json_path} is not a *.ast.json file"

        return cls._from_lean4_traced_file(root_dir, json_path, repo)

    @classmethod
    def _from_lean4_traced_file(
        cls, root_dir: Path, json_path: Path, repo: LeanGitRepo
    ) -> "TracedFile":
        lean_path = to_lean_path(root_dir, json_path, repo)
        lean_file = LeanFile(root_dir, lean_path)

        data = json.load(json_path.open())

        data["module_paths"] = []
        for line in (
            json_path.with_suffix("").with_suffix("").with_suffix(".dep_paths").open()
        ):
            line = line.strip()
            if line == "":
                break
            data["module_paths"].append(line)

        ast = FileNode.from_data(data, lean_file)
        comments = _collect_lean4_comments(ast)
        TracedFile._post_process_lean4(
            ast,
            lean_file,
            data["tactics"],
            data["premises"],
            data["module_paths"],
            comments,
        )

        return cls(root_dir, repo, lean_file, ast, comments)

    @classmethod
    def _post_process_lean4(
        cls,
        ast: FileNode,
        lean_file: LeanFile,
        tactics_data: List[Dict[str, Any]],
        premises_data: List[Dict[str, Any]],
        imports_data: List[str],
        comments: List[Comment],
    ) -> None:
        pos2tactics = {}
        for t in tactics_data:
            start = lean_file.position_to_pos(t["pos"])
            end = lean_file.position_to_pos(t["endPos"])
            pos2tactics[(start, end)] = t

        pos2premises = {}
        for p in premises_data:
            if (
                p is None
                or p["pos"] is None
                or p["endPos"] is None
                or p["fullName"] is None
                or p["fullName"] == "[anonymous]"
            ):
                continue
            start_line_nb, start_column_nb = p["pos"]["line"], p["pos"]["column"]
            end_line_nb, end_column_nb = p["endPos"]["line"], p["endPos"]["column"]
            start = Pos(line_nb=start_line_nb, column_nb=start_column_nb + 1)
            end = Pos(line_nb=end_line_nb, column_nb=end_column_nb + 1)
            pos2premises[(start, end)] = p

        inside_sections_namespaces = []

        def _callback(node: Node, _):
            if (
                isinstance(
                    node,
                    (
                        CommandNamespaceNode,
                        CommandSectionNode,
                        CommandNoncomputablesectionNode,
                    ),
                )
                and node.name is not None
            ):
                inside_sections_namespaces.append(node)
            elif (
                isinstance(node, CommandEndNode)
                and node.name is not None
                and len(inside_sections_namespaces) > 0
            ):
                inside_sections_namespaces.pop()
            elif is_potential_premise_lean4(node):
                prefix = ".".join(
                    ns.name
                    for ns in inside_sections_namespaces
                    if isinstance(ns, CommandNamespaceNode)
                )
                full_name = (
                    [_qualify_name(name, prefix) for name in node.name]
                    if is_mutual_lean4(node)
                    else _qualify_name(node.name, prefix)
                )
                object.__setattr__(node, "full_name", full_name)
                if isinstance(node, CommandDeclarationNode) and node.is_theorem:
                    object.__setattr__(node.get_theorem_node(), "full_name", full_name)
            elif isinstance(
                node,
                (
                    TacticTacticseq1IndentedNode,
                    TacticTacticseqbracketedNode,
                ),
            ):
                for tac_node in node.get_tactic_nodes():
                    assert isinstance(
                        tac_node, (OtherNode, TacticTacticseqbracketedNode)
                    )
                    if (tac_node.start, tac_node.end) not in pos2tactics:
                        continue
                    t = pos2tactics[(tac_node.start, tac_node.end)]
                    tac = get_code_without_comments(
                        lean_file, tac_node.start, tac_node.end, comments
                    )
                    tac = _fix_indentation(tac, tac_node.start.column_nb - 1)
                    object.__setattr__(tac_node, "state_before", t["stateBefore"])
                    object.__setattr__(tac_node, "state_after", t["stateAfter"])
                    object.__setattr__(tac_node, "tactic", tac)
            elif isinstance(node, IdentNode):
                start, end = node.get_closure()
                if (start, end) in pos2premises:
                    assert start is not None
                    assert end is not None
                    p = pos2premises[(start, end)]
                    prem = get_code_without_comments(lean_file, start, end, comments)
                    prem = _fix_indentation(prem, start.column_nb - 1)
                    if p["fullName"] is not None:
                        object.__setattr__(node, "full_name", p["fullName"])
                    if p["modName"] is not None:
                        object.__setattr__(node, "mod_name", p["modName"])
                    if p["defPath"] is not None:
                        object.__setattr__(node, "def_path", p["defPath"])
                    if p["defPos"] is not None and p["defEndPos"] is not None:
                        def_start_line_nb, def_start_column_nb = (
                            p["defPos"]["line"],
                            p["defPos"]["column"],
                        )
                        def_end_line_nb, def_end_column_nb = (
                            p["defEndPos"]["line"],
                            p["defEndPos"]["column"],
                        )
                        def_start = Pos(
                            line_nb=def_start_line_nb, column_nb=def_start_column_nb + 1
                        )
                        def_end = Pos(
                            line_nb=def_end_line_nb, column_nb=def_end_column_nb + 1
                        )
                        object.__setattr__(node, "def_start", def_start)
                        object.__setattr__(node, "def_end", def_end)
            elif isinstance(node, ModuleImportNode):
                node_module_name = object.__getattribute__(node, "module")
                if node_module_name is not None:
                    suffix = node_module_name.replace(".", "/")
                    for import_line in imports_data:
                        if import_line.endswith(
                            suffix + ".lean"
                        ) or import_line.endswith(suffix + "/default.lean"):
                            object.__setattr__(node, "path", Path(import_line))

        ast.traverse_preorder(_callback, node_cls=None)

    def check_sanity(self) -> None:
        """Perform some basic sanity checks.

        The function raises exceptions in case of unsuccessful checks.
        """
        assert isinstance(self.root_dir, Path)
        assert isinstance(self.lean_file, LeanFile)
        isinstance(self.ast, FileNode)

        assert self.lean_file.root_dir == self.root_dir

        for t in self.get_traced_theorems():
            assert str(self.lean_file.path).endswith(str(t.theorem.file_path))
            assert t.traced_file is None or t.traced_file is self

    def traverse_preorder(self, callback, node_cls: Optional[type] = None):
        """Traverse the AST in preorder.

        Args:
            callback (function): Callback function for visiting AST nodes.
            node_cls (Optional[type], optional): Restrict the application of
                ``callback`` to only nodes of type ``node_cls``.
                Defaults to None, which means applying ``callback`` to all.
        """
        self.ast.traverse_preorder(callback, node_cls)

    def _get_repo_and_relative_path(self) -> Tuple[LeanGitRepo, Path]:
        """Return the repo this file belongs to, as well as the file's path relative to it."""
        if self.path.is_relative_to(LEAN4_PACKAGES_DIR):
            # The theorem belongs to one of the dependencies.
            p = self.path.relative_to(LEAN4_PACKAGES_DIR)
            name = p.parts[0]
            repo = self.traced_repo.dependencies[name]
            return repo, p.relative_to(name)
        else:
            # The theorem belongs to the traced repo itself.
            return self.traced_repo.repo, self.path

    def get_traced_theorem(
        self, thm_or_name: Union[Theorem, str]
    ) -> Optional[TracedTheorem]:
        """Return a :class:`TracedTheorem` object given an :class:`Theorem` object
        or its fully-qualified name."""
        if isinstance(thm_or_name, Theorem):
            thm = thm_or_name
        else:
            repo, path = self._get_repo_and_relative_path()
            thm = Theorem(repo, path, thm_or_name)
        result = None
        private_result = None

        def _callback(
            node: Union[CommandTheoremNode, LemmaNode, MathlibTacticLemmaNode], _
        ) -> None:
            nonlocal result, private_result
            if not isinstance(
                node,
                (
                    CommandTheoremNode,
                    LemmaNode,
                    MathlibTacticLemmaNode,
                ),
            ):
                return False
            if node.full_name == thm.full_name:
                comments = self._filter_comments(node.start, node.end)
                t = TracedTheorem(self.root_dir, thm, node, comments, self)
                if t.is_private:
                    private_result = t
                else:
                    result = t

        self.ast.traverse_preorder(_callback, node_cls=None)

        # Prioritize non-private theorems.
        if result is None:
            result = private_result
        return result

    def get_traced_theorems(self) -> List[TracedTheorem]:
        """Return a list of traced theorem in this traced file."""
        traced_theorems = []

        def _callback(
            node: Union[CommandTheoremNode, LemmaNode, MathlibTacticLemmaNode], _
        ) -> None:
            if not isinstance(
                node,
                (
                    CommandTheoremNode,
                    LemmaNode,
                    MathlibTacticLemmaNode,
                ),
            ):
                return False
            repo, path = self._get_repo_and_relative_path()
            thm = Theorem(repo, path, node.full_name)
            comments = self._filter_comments(node.start, node.end)
            traced_theorems.append(
                TracedTheorem(self.root_dir, thm, node, comments, self)
            )
            # No need to traverse the subtree since theorems cannot be nested.
            return True

        self.traverse_preorder(_callback, node_cls=None)
        return traced_theorems

    def _filter_comments(self, start: Pos, end: Pos) -> List[Comment]:
        """Return a list of comments that are contained in the given range."""
        comments = []
        for c in self.comments:
            if c.start < start:
                assert c.end <= start
            elif c.start < end:
                assert c.end <= end
                comments.append(c)
        return comments

    def get_direct_dependencies(self, repo: LeanGitRepo) -> List[Tuple[str, Path]]:
        """Return the names and paths of all modules imported by the current :file:`*.lean` file."""
        deps = set()

        if not self.has_prelude:  # Add the prelude as a dependency.
            init_lean = Path("src/lean/Init.lean")
            if self.root_dir.name == "lean4":
                deps.add(("Init", init_lean))
            else:
                deps.add(("Init", LEAN4_PACKAGES_DIR / "lean4" / init_lean))

        def _callback(node: ModuleImportNode, _) -> None:
            if node.module is not None and node.path is not None:
                deps.add((node.module, node.path))

        self.traverse_preorder(_callback, node_cls=ModuleImportNode)
        return list(deps)

    def get_premise_definitions(self) -> List[Dict[str, Any]]:
        """Return all theorems and definitions defined in the current file that
        can be potentially used as premises.

        Returns:
            List[Dict[str, Any]]: _description_
        """
        results = []

        def _callback4(node: Node, _) -> None:
            if is_potential_premise_lean4(node):
                try:
                    start, end = node.get_closure()
                    if isinstance(node, CommandDeclarationNode) and node.is_theorem:
                        # We assume theorems are defined using keywords "theorem"
                        # or "lemma" but not, e.g., "def".
                        proof_start, _ = (
                            node.get_theorem_node().get_proof_node().get_closure()
                        )
                        code = get_code_without_comments(
                            self.lean_file, start, proof_start, self.comments
                        )
                        if code.endswith(":="):
                            code = code[:-2].strip()
                    else:
                        code = get_code_without_comments(
                            self.lean_file, start, end, self.comments
                        )
                    # TODO: For alias, restate_axiom, etc., the code is not very informative.
                    if is_mutual_lean4(node):
                        for s in node.full_name:
                            results.append(
                                {
                                    "full_name": s,
                                    "code": code,
                                    "start": list(start),
                                    "end": list(end),
                                    "kind": node.kind(),
                                }
                            )
                    else:
                        results.append(
                            {
                                "full_name": node.full_name,
                                "code": code,
                                "start": list(start),
                                "end": list(end),
                                "kind": node.kind(),
                            }
                        )
                except Exception as e:
                    logger.warning(f"Failed to process node: {e}")
                    logger.warning(f"Node: {node}")

        self.traverse_preorder(_callback4, node_cls=None)
        return results

    def to_xml(self) -> str:
        """Serialize a :class:`TracedFile` object to XML."""
        tree = etree.Element(self.__class__.__name__)

        tree.set("path", str(self.path))
        tree.set("md5", compute_md5(self.abs_path))

        self.ast.to_xml(tree)

        if self.comments is not None:
            comments_node = etree.SubElement(tree, "Comments")
            for c in self.comments:
                c.to_xml(comments_node)

        return etree.tostring(tree, encoding="utf-8", pretty_print=True).decode()

    @classmethod
    def from_xml(
        cls,
        root_dir: Union[str, Path],
        path: Union[str, Path],
        repo: LeanGitRepo,
    ) -> "TracedFile":
        """Load a :class:`TracedFile` object from its :file:`*.trace.xml` file.

        Args:
            root_dir (Union[str, Path]): Root directory of the traced repo.
            path (Union[str, Path]): Path of the :file:`*.trace.xml` file relative to ``root_dir``.
            repo (LeanGitRepo): The repo to which the traced file belongs.
        """
        root_dir = Path(root_dir)
        path = Path(path)
        assert path.suffixes == [".trace", ".xml"]
        lean_path = to_lean_path(root_dir, path, repo)
        lean_file = LeanFile(root_dir, lean_path)

        tree = etree.parse(path).getroot()
        assert tree.tag == "TracedFile"
        assert tree.attrib["path"] == str(lean_path)
        assert tree.attrib["md5"] == compute_md5(lean_file.abs_path)

        ast_tree, comments_tree = list(tree)
        ast = FileNode.from_xml(ast_tree, lean_file)
        comments = [Comment.from_xml(c) for c in comments_tree]

        return cls(root_dir, repo, lean_file, ast, comments)


def _save_xml_to_disk(tf: TracedFile) -> None:
    xml_path = tf.root_dir / to_xml_path(tf.root_dir, tf.path, tf.repo)
    with xml_path.open("wt") as oup:
        oup.write(tf.to_xml())


def _build_dependency_graph(
    seed_files: List[TracedFile], root_dir: Path, repo: LeanGitRepo
) -> nx.DiGraph:
    G = nx.DiGraph()

    for tf in seed_files:
        tf_path_str = str(tf.path)
        assert not G.has_node(tf_path_str)
        G.add_node(tf_path_str, traced_file=tf)

    traced_files = seed_files.copy()
    i = 0

    while i < len(traced_files):
        tf = traced_files[i]
        tf_path_str = str(tf.path)

        for dep_module, dep_path in tf.get_direct_dependencies(repo):
            try:
                dep_path_str = str(dep_path)
                if not G.has_node(dep_path_str):
                    json_path = to_json_path(root_dir, dep_path, repo)
                    tf_dep = TracedFile.from_traced_file(root_dir, json_path, repo)
                    G.add_node(dep_path_str, traced_file=tf_dep)
                    traced_files.append(tf_dep)

                G.add_edge(tf_path_str, dep_path_str, module=dep_module)
            except Exception as e:
                logger.warning(f"Failed to load dependency {dep_path}: {e}")
                continue

        i += 1  # Validate the graph is still acyclic, but don't fail if it's not
    assert nx.is_directed_acyclic_graph(G)

    return G


@dataclass(frozen=True, eq=False)
class TracedRepo:
    """A traced repo is a Lean repo of traced files and additional information, such as
    other repos it depends on, as well as the dependency graph between files.
    """

    repo: LeanGitRepo
    """The corresponding Lean repo.
    """

    dependencies: Dict[str, LeanGitRepo]
    """Dictionary mapping the name of each dependency to a :class:`LeanGitRepo` object.
    """

    root_dir: Path
    """Root directory of the traced repo.
    """

    traced_files: List[TracedFile] = field(repr=False)
    """List of traced files in the repo."""

    traced_files_graph: Optional[nx.DiGraph] = field(repr=False)
    """Dependency graph between files in the repo.
    
    The graph is a DAG, and there is an edge from file :file:`X` to file :file:`Y`
    if and only if :file:`X` imports :file:`Y`
    """

    def __post_init__(self) -> None:
        assert self.root_dir.is_absolute()

    def __setstate__(self, state) -> None:
        object.__setattr__(self, "__dict__", state)
        self._update_traced_files()

    @property
    def name(self) -> str:
        """Name of the repo."""
        return self.repo.name

    def show(self) -> None:
        """Show the repo in the default browser."""
        self.repo.show()

    def check_sanity(self) -> None:
        """Perform some basic sanity checks.

        The function raises exceptions in case of unsuccessful checks.
        """
        assert isinstance(self.repo, LeanGitRepo)
        assert isinstance(self.dependencies, dict)
        for k, v in self.dependencies.items():
            assert isinstance(k, str) and isinstance(v, LeanGitRepo)
        assert isinstance(self.root_dir, Path)
        assert self.traced_files_graph is None or isinstance(
            self.traced_files_graph, nx.DiGraph
        )

        assert self.repo not in self.dependencies.values()

        json_files = {
            p.relative_to(self.root_dir) for p in self.root_dir.glob("**/*.ast.json")
        }
        lean_files = {
            p.relative_to(self.root_dir) for p in self.root_dir.glob("**/*.lean")
        }
        xml_files = {
            p.relative_to(self.root_dir) for p in self.root_dir.glob("**/*.trace.xml")
        }
        path_files = {
            p.relative_to(self.root_dir) for p in self.root_dir.glob("**/*.dep_paths")
        }

        if self.traced_files_graph is not None:
            if not LOAD_USED_PACKAGES_ONLY:
                assert len(json_files) == self.traced_files_graph.number_of_nodes()

            for path_str, tf_node in self.traced_files_graph.nodes.items():
                tf = tf_node["traced_file"]
                path = Path(path_str)
                tf.check_sanity()
                assert tf.path == path and tf.root_dir == self.root_dir
                assert tf.traced_repo is None or tf.traced_repo is self
                assert path in lean_files
                assert (
                    to_dep_path(self.root_dir, path, self.repo) in path_files
                ), to_dep_path(self.root_dir, path, self.repo)
                assert (
                    to_json_path(self.root_dir, path, self.repo) in json_files
                ), to_json_path(self.root_dir, path, self.repo)
                if len(xml_files) > 0:
                    assert (
                        to_xml_path(self.root_dir, path, self.repo) in xml_files
                    ), to_xml_path(self.root_dir, path, self.repo)

    @classmethod
    def from_traced_files(
        cls, root_dir: Union[str, Path], build_deps: bool = True
    ) -> "TracedRepo":
        """Construct a :class:`TracedRepo` object by parsing :file:`*.ast.json` and :file:`*.path` files
           produced by :code:`lean --ast --tsast --tspp` (Lean 3) or :file:`ExtractData.lean` (Lean 4).

        Args:
            root_dir (Union[str, Path]): Root directory of the traced repo.
            build_deps (bool, optional): Whether to build the dependency graph between files.
        """
        root_dir = Path(root_dir).resolve()
        if not is_git_repo(root_dir):
            raise RuntimeError(f"{root_dir} is not a Git repo.")
        repo = LeanGitRepo.from_path(root_dir)

        json_paths = list(root_dir.glob("**/*.ast.json"))
        random.shuffle(json_paths)
        logger.debug(
            f"Parsing {len(json_paths)} *.ast.json files in {root_dir} with {NUM_WORKERS} workers"
        )

        traced_files = [
            TracedFile.from_traced_file(root_dir, path, repo)
            for path in tqdm(json_paths)
        ]

        dependencies = repo.get_dependencies(root_dir)
        if build_deps:
            traced_files_graph = _build_dependency_graph(traced_files, root_dir, repo)
        else:
            traced_files_graph = None

        traced_repo = cls(
            repo, dependencies, root_dir, traced_files, traced_files_graph
        )
        traced_repo._update_traced_files()
        return traced_repo

    def get_traced_file(self, path: Union[str, Path]) -> TracedFile:
        """Return a traced file by its path."""
        return self.traced_files_graph.nodes[str(path)]["traced_file"]

    def _update_traced_files(self) -> None:
        for tf in self.traced_files:
            tf.traced_repo = self

    def save_to_disk(self) -> None:
        """Save all traced files in the repo to the disk as :file:`*.trace.xml` files."""
        num_traced_files = len(self.traced_files)
        logger.debug(
            f"Saving {num_traced_files} traced XML files to {self.root_dir} with {NUM_WORKERS} workers"
        )

        for tf in tqdm(self.traced_files, total=num_traced_files):
            _save_xml_to_disk(tf)

    @classmethod
    def load_from_disk(
        cls, root_dir: Union[str, Path], build_deps: bool = True
    ) -> "TracedRepo":
        """Load a traced repo from :file:`*.trace.xml` files."""
        root_dir = Path(root_dir).resolve()
        if not is_git_repo(root_dir):
            raise RuntimeError(f"{root_dir} is not a Git repo.")
        repo = LeanGitRepo.from_path(root_dir)

        xml_paths = list(root_dir.glob("**/*.trace.xml"))
        logger.debug(
            f"Loading {len(xml_paths)} traced XML files from {root_dir} with {NUM_WORKERS} workers"
        )

        # Start from files in the target repo as seeds.
        # Only load dependency files that are actually used.
        if LOAD_USED_PACKAGES_ONLY:
            xml_paths = [
                p
                for p in xml_paths
                if not "lake-packages/" in str(p) and not ".lake/packages" in str(p)
            ]

        traced_files = [
            TracedFile.from_xml(root_dir, path, repo) for path in tqdm(xml_paths)
        ]

        dependencies = repo.get_dependencies(root_dir)
        if build_deps:
            traced_files_graph = _build_dependency_graph(traced_files, root_dir, repo)
        else:
            traced_files_graph = None

        traced_repo = cls(
            repo, dependencies, root_dir, traced_files, traced_files_graph
        )
        traced_repo._update_traced_files()
        return traced_repo

    def get_traced_theorems(self) -> List[TracedTheorem]:
        """Return all traced theorems in the repo."""
        return list(
            itertools.chain.from_iterable(
                tf.get_traced_theorems() for tf in self.traced_files
            )
        )

    def get_traced_theorem(self, thm: Theorem) -> Optional[TracedTheorem]:
        """Return a :class:`TracedTheorem` object corresponding to ``thm``"""
        if thm.repo == self.repo:
            path = Path(thm.repo.name) / thm.file_path
        else:
            assert thm.repo in self.dependencies.values()
            path = Path(self.name) / LEAN4_PACKAGES_DIR / thm.repo.name / thm.file_path
        return self.get_traced_file(path).get_traced_theorem(thm.full_name)



import os
from typing import List, Tuple

import requests
from loguru import logger

from .external_parser import *


class HFTacticGenerator(Generator, Transformer):
    def __init__(
        self, model_name: str = "deepseek-ai/DeepSeek-Prover-V2-671B:novita", **args
    ) -> None:
        self.model_name = model_name
        self.used_tactics = set()

    def generate(self, input: str, target_prefix: str = "") -> List[Tuple[str, float]]:
        prompt = input + target_prefix
        '''prompt= 'Here is a theorem you need to prove in Lean:\n'+prompt+'\nNow you should suggest one line tactic in lean code:'
        prompt = f"""<|im_start|>user\n{prompt}<|im_end|>\n<|im_start|>assistant\n"""
        '''
        prompt = pre_process_input(self.model_name, prompt)

        API_URL = "https://router.huggingface.co/v1/chat/completions"
        headers = {
            "Authorization": f"Bearer {os.environ['HF_TOKEN']}",
        }

        payload = {
            "messages": [{"role": "user", "content": prompt}],
            "model": "deepseek-ai/DeepSeek-Prover-V2-671B:novita",
            "temperature": 0.6,
            "max_tokens": 256,
            "top_p": 0.9,
            "n": 4,
        }

        response_data = requests.post(API_URL, headers=headers, json=payload).json()
        response = [
            choice["message"]["content"] for choice in response_data.get("choices", [])
        ]

        if response_data.get("error"):
            logger.error(f"Error generating tactic: {response_data['error']}")
            return "sorry"

        result = []
        for out in response:
            out = post_process_output(self.model_name, out)
            result.append(out)

        for r in result:
            if (input, r) not in self.used_tactics:
                self.used_tactics.add((input, r))
                logger.info(f"tactic: {r}")
                return r

        return "sorry"

    def generate_whole_proof(self, theorem: str) -> str:
        prompt = (
            "Here is a theorem you need to prove in Lean:```lean\n"
            + theorem
            + "```\nNow you should provide the complete proof of the theorem in lean code. Only output the proof, no explanation, no comments, no theorem, nothing else:"
        )

        API_URL = "https://router.huggingface.co/v1/chat/completions"
        headers = {
            "Authorization": f"Bearer {os.environ['HF_TOKEN']}",
        }

        payload = {
            "messages": [{"role": "user", "content": prompt}],
            "model": "deepseek-ai/DeepSeek-Prover-V2-671B:novita",
            "temperature": 0.6,
            "max_tokens": 256,
            "top_p": 0.9,
            "n": 1,
        }

        response_data = requests.post(API_URL, headers=headers, json=payload).json()
        response = [
            choice["message"]["content"] for choice in response_data.get("choices", [])
        ]

        if response_data.get("error"):
            logger.error(f"Error generating tactic: {response_data['error']}")
            return "sorry"

        if len(response) == 0:
            return "sorry"

        out = post_process_output(self.model_name, response[0])

        return out


if __name__ == "__main__":
    model = HFTacticGenerator()
    print(model.generate("n : ℕ\n⊢ gcd n n = n"))



"""Trainer package for Lean theorem proving models."""

from .grpo_trainer import GRPOTrainer
from .progress_trainer import ProgressTrainer
from .retrieval_trainer import RetrievalTrainer
from .sft_trainer import SFTTrainer

__all__ = ["RetrievalTrainer", "SFTTrainer", "GRPOTrainer", "ProgressTrainer"]



__version__ = "1.0.0"
__author__ = "LeanDojo-v2 Contributors"

from .config import ProverConfig, TrainingConfig
from .database.dynamic_database import DynamicDatabase

__all__ = [
    "DynamicDatabase",
    "TrainingConfig",
    "ProverConfig",
]



from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np
import torch


def get_cuda_if_available():
    return torch.device("cuda" if torch.cuda.is_available() else "cpu")


def pre_process_input(model_name, input):
    if (
        model_name == "internlm/internlm2-math-plus-1_8b"
        or model_name == "AI-MO/Kimina-Prover-Preview-Distill-7B"
    ):
        prompt = (
            "My LEAN 4 state is:\n```lean\n"
            + input
            + "```\nPlease predict a possible tactic to help me prove the theorem."
        )
        prompt = f"""<|im_start|>user\n{prompt}<|im_end|>\n<|im_start|>assistant\n"""
    elif model_name == "gpt-3.5-turbo" or model_name == "gpt-4-turbo-preview":
        prompt = (
            "Here is a theorem you need to prove in Lean:\n"
            + input
            + "\nNow you should suggest one line tactic in lean code:"
        )
    elif (
        model_name == "deepseek-ai/DeepSeek-Prover-V2-7B"
        or model_name == "deepseek-ai/DeepSeek-Prover-V2-671B:novita"
    ):
        prompt = (
            "Here is a theorem you need to prove in Lean:```lean\n"
            + input
            + "```\nNow you should suggest exactly one line tactic in lean code. Only output the tactic, no explanation, no comments, no theorem, nothing else:"
        )
    elif "gemini" in model_name or "claude" in model_name:
        prompt = (
            "Here is a theorem you need to prove in Lean:\n"
            + input
            + "\nNow you should suggest one line tactic in lean code:"
        )
    else:
        raise NotImplementedError(f"External model '{model_name}' not supported")
    return prompt


def post_process_output(model_name, output):
    if model_name == "internlm/internlm2-math-plus-1_8b":
        result = (
            output.split("assistant")[-1]
            .split("lean")[-1]
            .split("```")[0]
            .split("\n")[1]
        )
    elif model_name == "AI-MO/Kimina-Prover-Preview-Distill-7B":
        result = (
            output.split("assistant")[-1]
            .split("lean")[-1]
            .split("```")[0]
            .split("\n")[-2]
            .lstrip()
        )
    elif model_name == "deepseek-ai/DeepSeek-Prover-V2-7B":
        result = output.split("lean")[-1].split("```")[0].split("\n")[2].lstrip()
    elif model_name == "deepseek-ai/DeepSeek-Prover-V2-671B:novita":
        result = output.split("lean")[-1].replace("`", "").replace("by", "").lstrip()
    elif model_name == "gpt-3.5-turbo" or model_name == "gpt-4-turbo-preview":
        result = output.split("lean")[-1].split("```")[0].split("\n")[1]
    elif "gemini" in model_name or "claude" in model_name:
        result = output.split("lean")[-1].split("```")[0].split("\n")[1]
    else:
        raise NotImplementedError(f"External model '{model_name}' not supported")
    return result


def choices_dedup(output_list: List[tuple[str, float]]) -> List[tuple[str, float]]:
    unique_data = {}
    for item in output_list:
        if item[0] not in unique_data or item[1] > unique_data[item[0]]:
            unique_data[item[0]] = item[1]
    sorted_data = sorted(unique_data.items(), key=lambda x: x[1], reverse=True)
    return sorted_data


class Generator(ABC):
    @abstractmethod
    def generate(self, input: str, target_prefix: str = "") -> List[Tuple[str, float]]:
        pass


class Encoder(ABC):
    @abstractmethod
    def encode(self, input: str) -> np.ndarray:
        pass


class Transformer:
    def cuda(self) -> None:
        self.model.cuda()

    def cpu(self) -> None:
        self.model.cpu()

    @property
    def device(self) -> torch.device:
        return self.model.device



"""RAG Trainer for Lean theorem proving models."""

import os
from pathlib import Path
from typing import List, Optional

import torch
from loguru import logger
from pytorch_lightning import seed_everything
from tqdm import tqdm

from lean_dojo_v2.database import DynamicDatabase
from lean_dojo_v2.database.models.repository import Repository
from lean_dojo_v2.lean_agent.config import TrainingConfig
from lean_dojo_v2.lean_agent.retrieval.datamodule import RetrievalDataModule
from lean_dojo_v2.lean_agent.retrieval.model import PremiseRetriever
from lean_dojo_v2.lean_dojo.data_extraction.lean import LeanGitRepo
from lean_dojo_v2.utils.constants import DATA_DIR, RAID_DIR
from lean_dojo_v2.utils.filesystem import find_latest_checkpoint


class RetrievalTrainer:
    """Trainer class for RAG-based theorem proving models.

    This class handles the training of retrieval-augmented generation models
    for Lean theorem proving, including model loading, data preparation,
    and training execution.
    """

    def __init__(self, config: Optional[TrainingConfig] = None):
        """Initialize the RAG trainer.

        Args:
            config: Training configuration. If None, uses default TrainingConfig.
        """
        self.config = config or TrainingConfig()

    def train(
        self,
        repos: List[Repository],
        database: DynamicDatabase,
        data_path: Path = Path(os.path.join(RAID_DIR, DATA_DIR, "merged")),
        model_checkpoint_path: Optional[str] = None,
    ) -> None:
        """Train the model on the specified repository.

        Args:
            repo_url: URL of the repository being trained on
            repo_commit: Commit hash of the repository
            model_checkpoint_path: Path to the model checkpoint to load from.
                                 If None, finds the latest checkpoint.
        """
        repos_to_process = []

        for repo in repos:
            repos_to_process.append(repo)

            database.export_merged_data(repos_to_process, data_path)

            if model_checkpoint_path is None:
                model_checkpoint_path = find_latest_checkpoint()
            logger.info(f"Found latest checkpoint: {model_checkpoint_path}")

            seed_everything(self.config.seed)

            if not torch.cuda.is_available():
                logger.warning("Indexing the corpus using CPU can be very slow.")
                device = torch.device("cpu")
            else:
                device = torch.device("cuda")

            model_config = self.config.get_model_config()
            model = PremiseRetriever.load(
                model_checkpoint_path, device, freeze=False, config=model_config
            )
            model.train()
            logger.info(f"Loaded premise retriever at {model_checkpoint_path}")

            # Create custom log directory
            custom_log_dir = self.config.get_custom_log_dir(str(repo))

            # Create trainer using configuration
            trainer = self.config.create_trainer(str(repo), custom_log_dir)

            model.set_lambda(self.config.get_lambda_value())

            corpus_path = os.path.join(data_path, "corpus.jsonl")
            random_path = os.path.join(data_path, "random")

            data_module_config = self.config.get_data_module_config(
                random_path, corpus_path
            )
            data_module = RetrievalDataModule(**data_module_config)
            data_module.setup(stage="fit")

            logger.info(f"Training dataset size: {len(data_module.ds_train)}")
            logger.info(f"Validation dataset size: {len(data_module.ds_val)}")
            logger.info(f"Testing dataset size: {len(data_module.ds_pred)}")

            trainer.fit(
                model,
                datamodule=data_module,
            )

    def evaluate(self, dataset_path: Optional[str] = None) -> None:
        """Evaluate the trained model.

        Args:
            dataset_path: Path to the dataset directory. If None, uses default RAID_DIR/DATA_DIR.
        """
        best_model_path = find_latest_checkpoint()
        logger.info("Testing...")

        if dataset_path is None:
            dataset_path = os.path.join(RAID_DIR, DATA_DIR)

        testing_paths = [
            os.path.join(dataset_path, d) for d in os.listdir(dataset_path)
        ]

        for data_path in testing_paths:
            if not os.path.isdir(data_path):
                continue

            # Import here to avoid circular imports
            from lean_dojo_v2.lean_agent.retrieval.main import run_cli

            run_cli(best_model_path, data_path)

            num_gpus = self.config.num_gpus
            preds_map = {}
            for gpu_id in range(num_gpus):
                pickle_path = f"test_pickle_{gpu_id}.pkl"
                if os.path.exists(pickle_path):
                    import pickle

                    with open(pickle_path, "rb") as f:
                        preds = pickle.load(f)
                        preds_map.update(preds)

            logger.info("Loaded the predictions pickle files")
            test_data_path = os.path.join(data_path, "random", "test.json")

            if os.path.exists(test_data_path):
                import json

                with open(test_data_path, "r") as f:
                    data = json.load(f)

                R1, R10, MRR = self._evaluate_retrieval_metrics(data, preds_map)
                logger.info(f"R@1: {R1:.2f}, R@10: {R10:.2f}, MRR: {MRR:.2f}")

    def _evaluate_retrieval_metrics(self, data, preds_map):
        """Evaluate retrieval metrics on the test data.

        Args:
            data: Test data
            preds_map: Predictions map

        Returns:
            Tuple of (R@1, R@10, MRR) metrics
        """
        R1 = []
        R10 = []
        MRR = []

        for thm in tqdm(data):
            for i, _ in enumerate(thm["traced_tactics"]):
                pred = None
                key = (thm["file_path"], thm["full_name"], tuple(thm["start"]), i)
                if key in preds_map:
                    pred = preds_map[key]
                else:
                    continue
                all_pos_premises = set(pred["all_pos_premises"])
                if len(all_pos_premises) == 0:
                    continue

                retrieved_premises = pred["retrieved_premises"]
                TP1 = retrieved_premises[0] in all_pos_premises
                R1.append(float(TP1) / len(all_pos_premises))
                TP10 = len(all_pos_premises.intersection(retrieved_premises[:10]))
                R10.append(float(TP10) / len(all_pos_premises))

                for j, p in enumerate(retrieved_premises):
                    if p in all_pos_premises:
                        MRR.append(1.0 / (j + 1))
                        break
                else:
                    MRR.append(0.0)

        R1 = 100 * sum(R1) / len(R1) if R1 else 0.0
        R10 = 100 * sum(R10) / len(R10) if R10 else 0.0
        MRR = sum(MRR) / len(MRR) if MRR else 0.0
        return R1, R10, MRR



from typing import Optional

from peft import LoraConfig

from lean_dojo_v2.lean_agent.config import ProverConfig, TrainingConfig
from lean_dojo_v2.prover import HFProver
from lean_dojo_v2.trainer import SFTTrainer
from lean_dojo_v2.utils.constants import *

from .base_agent import BaseAgent


class HFAgent(BaseAgent):
    def __init__(
        self,
        trainer: SFTTrainer,
        database_path: str = "dynamic_database.json",
        training_config: Optional[TrainingConfig] = None,
        prover_config: Optional[ProverConfig] = None,
    ):
        super().__init__(database_path)
        self.config = training_config or TrainingConfig()
        self.prover_config = prover_config or ProverConfig()
        self.trainer = trainer
        self.output_dir = self.trainer.output_dir
        self.use_lora = self.trainer.lora_config is not None

    def _get_build_deps(self) -> bool:
        """HFAgent doesn't build dependencies by default."""
        return False

    def _setup_prover(self):
        """Set up the HFProver for HFAgent."""
        self.prover = HFProver(
            ckpt_path=self.output_dir,
            use_lora=self.use_lora,
        )


def main():
    """
    Main function to run HFAgent.
    """
    url = "https://github.com/durant42040/lean4-example"
    commit = "005de00d03f1aaa32cb2923d5e3cbaf0b954a192"
    model_name = "deepseek-ai/DeepSeek-Prover-V2-7B"
    output_dir = "outputs-deepseek"

    lora_config = LoraConfig(
        r=16,
        lora_alpha=32,
        target_modules=["q_proj", "v_proj", "k_proj", "o_proj"],
        lora_dropout=0.1,
    )

    trainer = SFTTrainer(
        model_name=model_name,
        output_dir=output_dir,
        epochs_per_repo=1,
        batch_size=2,
        lr=2e-5,
        lora_config=lora_config,
    )

    agent = HFAgent(trainer=trainer)
    agent.setup_github_repository(url=url, commit=commit)
    agent.train()
    agent.prove()


if __name__ == "__main__":
    main()



import os
from typing import Optional

from lean_dojo_v2.lean_agent.config import ProverConfig, TrainingConfig
from lean_dojo_v2.prover import RetrievalProver
from lean_dojo_v2.trainer import RetrievalTrainer
from lean_dojo_v2.utils.constants import RAID_DIR
from lean_dojo_v2.utils.filesystem import find_latest_checkpoint

from .base_agent import BaseAgent


class LeanAgent(BaseAgent):
    def __init__(
        self,
        database_path: str = "dynamic_database.json",
        training_config: Optional[TrainingConfig] = None,
        prover_config: Optional[ProverConfig] = None,
    ):
        super().__init__(database_path)
        self.config = training_config or TrainingConfig()
        self.prover_config = prover_config or ProverConfig()
        self.trainer = RetrievalTrainer(config=self.config)

    def _get_build_deps(self) -> bool:
        """LeanAgent builds dependencies by default."""
        return True

    def _setup_prover(self):
        """Set up the RetrievalProver for LeanAgent."""
        ret_ckpt_path = find_latest_checkpoint()
        self.prover = RetrievalProver(
            ret_ckpt_path=ret_ckpt_path,
            gen_ckpt_path=os.path.join(RAID_DIR, "model_lightning.ckpt"),
            indexed_corpus_path=os.path.join(self.data_path, "corpus.jsonl"),
        )


def main():
    """
    Main function to run LeanAgent.
    """
    url = "https://github.com/durant42040/lean4-example"
    commit = "005de00d03f1aaa32cb2923d5e3cbaf0b954a192"

    agent = LeanAgent()
    agent.setup_github_repository(url=url, commit=commit)
    agent.train()
    agent.prove()


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np
import torch
from loguru import logger
from transformers import (
    AutoModelForCausalLM,
    AutoModelForSeq2SeqLM,
    AutoModelForTextEncoding,
    AutoTokenizer,
)


class Generator(ABC):
    @abstractmethod
    def generate(self, input: str, target_prefix: str = "") -> List[Tuple[str, float]]:
        pass


class Encoder(ABC):
    @abstractmethod
    def encode(self, input: str) -> np.ndarray:
        pass


class Transformer:
    def cuda(self) -> None:
        self.model.cuda()

    def cpu(self) -> None:
        self.model.cpu()

    @property
    def device(self) -> torch.device:
        return self.model.device


def get_cuda_if_available():
    return torch.device("cuda" if torch.cuda.is_available() else "cpu")


class DecoderOnlyTransformer(Generator, Transformer):
    def __init__(
        self,
        name: str,
        num_return_sequences: int,
        max_length: int,
        length_penalty: float = 0.0,
        device: str = "cpu",
    ) -> None:
        self.tokenizer = AutoTokenizer.from_pretrained(name)
        if device == "auto":
            device = get_cuda_if_available()
        else:
            device = torch.device(device)
        logger.info(f"Loading {name} on {device}")
        self.model = AutoModelForCausalLM.from_pretrained(name).to(device)
        self.max_length = max_length
        self.num_return_sequences = num_return_sequences
        self.length_penalty = length_penalty

    def generate(self, input: str, target_prefix: str = "") -> List[Tuple[str, float]]:
        tokenized_input = self.tokenizer(input + target_prefix, return_tensors="pt")
        output = self.model.generate(
            tokenized_input.input_ids.to(self.device),
            max_length=self.max_length,
            num_beams=self.num_return_sequences,
            length_penalty=self.length_penalty,
            do_sample=False,
            num_return_sequences=self.num_return_sequences,
            early_stopping=False,
            return_dict_in_generate=True,
            output_scores=True,
        )
        raw_outputs = self.tokenizer.batch_decode(
            output.sequences, skip_special_tokens=True
        )
        outputs = []

        for out, score in zip(raw_outputs, output.sequences_scores.exp()):
            assert out.startswith(input + target_prefix)
            outputs.append((out[len(input) :], score.item()))

        return outputs


class PythiaTacticGenerator(DecoderOnlyTransformer):
    def __init__(
        self,
        num_return_sequences: int,
        max_length: int,
        length_penalty: float = 0.0,
        device: str = "cpu",
    ) -> None:
        super().__init__(
            "wellecks/llmstep-mathlib4-pythia2.8b",
            num_return_sequences,
            max_length,
            length_penalty,
            device,
        )

    def generate(self, input: str, target_prefix: str = "") -> List[Tuple[str, float]]:
        return super().generate(f"[GOAL]{input}[PROOFSTEP]{target_prefix}")


class EncoderDecoderTransformer(Generator, Transformer):
    def __init__(
        self,
        name: str,
        num_return_sequences: int,
        max_length: int,
        length_penalty: float = 0.0,
        device: str = "cpu",
    ) -> None:
        self.tokenizer = AutoTokenizer.from_pretrained(name)
        if device == "auto":
            device = get_cuda_if_available()
        else:
            device = torch.device(device)
        logger.info(f"Loading {name} on {device}")
        self.model = AutoModelForSeq2SeqLM.from_pretrained(name)
        self.max_length = max_length
        self.num_return_sequences = num_return_sequences
        self.length_penalty = length_penalty

    def generate(self, input: str, target_prefix: str = "") -> List[Tuple[str, float]]:
        assert (
            target_prefix == ""
        ), "target_prefix is not supported by encoder-decoder Transformer"
        tokenized_input = self.tokenizer(input, return_tensors="pt")
        output = self.model.generate(
            tokenized_input.input_ids.to(self.device),
            max_length=self.max_length,
            num_beams=self.num_return_sequences,
            length_penalty=self.length_penalty,
            do_sample=False,
            num_return_sequences=self.num_return_sequences,
            early_stopping=False,
            return_dict_in_generate=True,
            output_scores=True,
        )
        raw_outputs = self.tokenizer.batch_decode(
            output.sequences, skip_special_tokens=True
        )
        return list(zip(raw_outputs, output.sequences_scores.exp().tolist()))


class EncoderOnlyTransformer(Encoder, Transformer):
    def __init__(self, name: str, device: str = "cpu") -> None:
        self.tokenizer = AutoTokenizer.from_pretrained(name)
        if device == "auto":
            device = get_cuda_if_available()
        else:
            device = torch.device(device)
        logger.info(f"Loading {name} on {device}")
        self.model = AutoModelForTextEncoding.from_pretrained(name)

    @torch.no_grad()
    def encode(self, input: str) -> np.ndarray:
        tokenized_input = self.tokenizer(input, return_tensors="pt")
        hidden_state = self.model(
            tokenized_input.input_ids.to(self.device)
        ).last_hidden_state
        feature = hidden_state.mean(dim=1).squeeze()
        return feature.cpu().numpy()


if __name__ == "__main__":
    model = PythiaTacticGenerator(num_return_sequences=32, max_length=1024)
    model.cuda()
    print(model.generate("n : ℕ\n⊢ gcd n n = n"))



from .dynamic_database import DynamicDatabase
from .models import (
    AnnotatedTactic,
    Annotation,
    Premise,
    PremiseFile,
    Repository,
    Theorem,
)

__all__ = [
    "DynamicDatabase",
    "Repository",
    "Theorem",
    "Annotation",
    "AnnotatedTactic",
    "Premise",
    "PremiseFile",
]



from lean_dojo_v2.utils.constants import __version__

from .data_extraction.dataset import generate_benchmark
from .data_extraction.lean import LeanFile, LeanGitRepo, Pos, Theorem, get_latest_commit
from .data_extraction.trace import get_traced_repo_path, is_available_in_cache, trace
from .data_extraction.traced_data import (
    TracedFile,
    TracedRepo,
    TracedTactic,
    TracedTheorem,
)



from __future__ import annotations

import datetime
import json
import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional

from tqdm import tqdm

from .premises import PremiseFile
from .theorems import Theorem


@dataclass
class Repository:
    """
    Repository class represents a repository with various attributes and methods
    to manage theorems and premise files.

    Attributes:
        url (str): URL of the repository.
        name (str): Name of the repository.
        commit (str): Commit hash of the repository.
        lean_version (str): Version of Lean used in the repository.
        lean_dojo_version (str): Version of Lean Dojo used in the repository.
        metadata (Dict[str, str]): Metadata associated with the repository.
        proven_theorems (List[Theorem]): List of proven theorems.
        sorry_theorems_proved (List[Theorem]): List of sorry theorems that have been proved.
        sorry_theorems_unproved (List[Theorem]): List of sorry theorems that are unproved.
        premise_files (List[PremiseFile]): List of premise files.
        files_traced (List[Path]): List of traced files.
        pr_url (Optional[str]): URL of the pull request.
    """

    url: str
    name: str
    commit: str
    lean_version: str
    lean_dojo_version: str
    metadata: Dict[str, str]
    proven_theorems: List[Theorem] = field(default_factory=list)
    sorry_theorems_proved: List[Theorem] = field(default_factory=list)
    sorry_theorems_unproved: List[Theorem] = field(default_factory=list)
    premise_files: List[PremiseFile] = field(default_factory=list)
    files_traced: List[Path] = field(default_factory=list)
    pr_url: Optional[str] = None

    def __eq__(self, other):
        if not isinstance(other, Repository):
            return NotImplemented
        return (
            self.url == other.url
            and self.name == other.name
            and self.commit == other.commit
            and self.lean_version == other.lean_version
            and self.lean_dojo_version == other.lean_dojo_version
        )

    def __hash__(self):
        return hash(
            (
                self.url,
                self.name,
                self.commit,
                self.lean_version,
                self.lean_dojo_version,
            )
        )

    def __str__(self) -> str:
        """Directory-friendly string representation.

        Example: "repo-name_commit-hash"
        """
        return f"{self.name}_{self.commit}"

    @property
    def total_theorems(self) -> int:
        return self.num_proven_theorems + self.num_sorry_theorems

    @property
    def num_proven_theorems(self) -> int:
        return len(self.proven_theorems)

    @property
    def num_sorry_theorems_proved(self) -> int:
        return len(self.sorry_theorems_proved)

    @property
    def num_sorry_theorems_unproved(self) -> int:
        return len(self.sorry_theorems_unproved)

    @property
    def num_sorry_theorems(self) -> int:
        return self.num_sorry_theorems_proved + self.num_sorry_theorems_unproved

    @property
    def num_premise_files(self) -> int:
        return len(self.premise_files)

    @property
    def num_premises(self) -> int:
        return sum(len(pf.premises) for pf in self.premise_files)

    @property
    def num_files_traced(self) -> int:
        return len(self.files_traced)

    @property
    def get_all_theorems(self) -> List[Theorem]:
        return (
            self.proven_theorems
            + self.sorry_theorems_proved
            + self.sorry_theorems_unproved
        )

    def get_theorem(self, full_name: str, file_path: str) -> Optional[Theorem]:
        for thm_list in [
            self.proven_theorems,
            self.sorry_theorems_proved,
            self.sorry_theorems_unproved,
        ]:
            for thm in thm_list:
                if thm.full_name == full_name and (
                    str(thm.file_path) == file_path
                    or (file_path == "" and str(thm.file_path) == ".")
                ):
                    return thm
        return None

    def update_theorem(self, theorem: Theorem) -> None:
        for thm_list in [
            self.proven_theorems,
            self.sorry_theorems_proved,
            self.sorry_theorems_unproved,
        ]:
            for i, thm in enumerate(thm_list):
                if thm.is_same_theorem(theorem):
                    thm_list[i] = theorem
                    return
        raise ValueError(f"Theorem '{theorem.full_name}' not found.")

    def get_premise_file(self, path: str) -> Optional[PremiseFile]:
        return next((pf for pf in self.premise_files if str(pf.path) == path), None)

    def get_file_traced(self, path: str) -> Optional[Path]:
        return next((f for f in self.files_traced if str(f) == path), None)

    @classmethod
    def from_dict(cls, data: Dict) -> Repository:
        if not all(
            key in data
            for key in [
                "url",
                "name",
                "commit",
                "lean_version",
                "lean_dojo_version",
                "metadata",
            ]
        ):
            raise ValueError("Invalid Repository data format")
        if "date_processed" not in data["metadata"]:
            raise ValueError("Metadata must contain the 'date_processed' key")

        metadata = data["metadata"].copy()
        if isinstance(metadata["date_processed"], str):
            metadata["date_processed"] = datetime.datetime.fromisoformat(
                metadata["date_processed"]
            )

        repo = cls(
            url=data["url"],
            name=data["name"],
            commit=data["commit"],
            lean_version=data["lean_version"],
            lean_dojo_version=data["lean_dojo_version"],
            metadata=metadata,
            files_traced=[],
            pr_url=data.get("pr_url"),
        )

        if all(
            key in data
            for key in ["theorems_folder", "premise_files_corpus", "files_traced"]
        ):
            if not all(
                os.path.exists(data[key])
                for key in ["theorems_folder", "premise_files_corpus", "files_traced"]
            ):
                raise ValueError(
                    "Paths to data cannot be empty when creating repo from dataset"
                )

            theorems_folder = Path(data["theorems_folder"])
            for file in theorems_folder.glob("*.json"):
                with open(file, "r") as f:
                    theorem_data = json.load(f)
                for t_data in tqdm(theorem_data):
                    theorem = Theorem.from_dict(t_data, repo.url, repo.commit)
                    if any(
                        "sorry" in step.tactic
                        for step in (theorem.traced_tactics or [])
                    ):
                        repo.sorry_theorems_unproved.append(theorem)
                    else:
                        repo.proven_theorems.append(theorem)

            with open(data["premise_files_corpus"], "r") as f:
                for line in f:
                    premise_file_data = json.loads(line)
                    premise_file = PremiseFile.from_dict(premise_file_data)
                    repo.premise_files.append(premise_file)

            with open(data["files_traced"], "r") as f:
                for line in f:
                    traced_file_data = json.loads(line)
                    repo.files_traced.append(Path(traced_file_data["traced_file_path"]))
        else:
            # Process theorems and premises from the existing data structure
            repo.proven_theorems = [
                Theorem.from_dict(t, repo.url, repo.commit)
                for t in data.get("proven_theorems", [])
            ]
            repo.sorry_theorems_proved = [
                Theorem.from_dict(t, repo.url, repo.commit)
                for t in data.get("sorry_theorems_proved", [])
            ]
            repo.sorry_theorems_unproved = [
                Theorem.from_dict(t, repo.url, repo.commit)
                for t in data.get("sorry_theorems_unproved", [])
            ]
            repo.premise_files = [
                PremiseFile.from_dict(pf) for pf in data.get("premise_files", [])
            ]
            repo.files_traced = [Path(file) for file in data.get("files_traced", [])]

        return repo

    def to_dict(self) -> Dict:
        metadata_copy = self.metadata.copy()
        if isinstance(metadata_copy["date_processed"], datetime.datetime):
            metadata_copy["date_processed"] = metadata_copy[
                "date_processed"
            ].isoformat()
        return {
            "url": self.url,
            "name": self.name,
            "commit": self.commit,
            "lean_version": self.lean_version,
            "lean_dojo_version": self.lean_dojo_version,
            "metadata": metadata_copy,
            "total_theorems": self.total_theorems,
            "num_proven_theorems": self.num_proven_theorems,
            "num_sorry_theorems": self.num_sorry_theorems,
            "num_sorry_theorems_proved": self.num_sorry_theorems_proved,
            "num_sorry_theorems_unproved": self.num_sorry_theorems_unproved,
            "num_premise_files": self.num_premise_files,
            "num_premises": self.num_premises,
            "num_files_traced": self.num_files_traced,
            "proven_theorems": [t.to_dict() for t in self.proven_theorems],
            "sorry_theorems_proved": [t.to_dict() for t in self.sorry_theorems_proved],
            "sorry_theorems_unproved": [
                t.to_dict() for t in self.sorry_theorems_unproved
            ],
            "premise_files": [pf.to_dict() for pf in self.premise_files],
            "files_traced": [str(file) for file in self.files_traced],
            "pr_url": self.pr_url,
        }

    def change_sorry_to_proven(self, theorem: Theorem, log_file: str) -> None:
        if theorem in self.sorry_theorems_unproved:
            self.sorry_theorems_unproved.remove(theorem)
            self.sorry_theorems_proved.append(theorem)

            message = f"Theorem proved: {theorem.full_name} in {theorem.file_path} for repo {self.name} (commit: {self.commit})"
            timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            log_entry = f"{timestamp} - {message}\n"

            os.makedirs(os.path.dirname(log_file), exist_ok=True)

            with open(log_file, "a") as f:
                f.write(log_entry)
        else:
            raise ValueError(
                "The theorem is not in the list of unproved sorry theorems."
            )



"""RAG agent for Lean theorem proving."""

import random
from typing import Optional

import torch
from pantograph.expr import GoalState, Tactic
from peft import AutoPeftModelForCausalLM
from transformers import AutoModelForCausalLM, AutoTokenizer

from lean_dojo_v2.database.models.theorems import Theorem
from lean_dojo_v2.prover.base_prover import BaseProver


class HFProver(BaseProver):
    """Retrieval-Augmented Generation agent for Lean theorem proving.

    This agent uses a retrieval-augmented generator to generate tactics
    by first retrieving relevant premises and then generating tactics
    based on the retrieved context.
    """

    def __init__(self, ckpt_path: str, use_lora: bool = False, device: str = "auto"):
        super().__init__()
        if device == "auto":
            self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            self.device = torch.device(device)

        self.tokenizer = AutoTokenizer.from_pretrained(ckpt_path)

        if use_lora:
            self.model = AutoPeftModelForCausalLM.from_pretrained(ckpt_path).to(
                self.device
            )
        else:
            self.model = AutoModelForCausalLM.from_pretrained(ckpt_path).to(self.device)

        self.model.eval()

    def next_tactic(
        self,
        state: GoalState,
        goal_id: int,
    ) -> Optional[Tactic]:
        """Generate the next tactic using the loaded HuggingFace model."""
        if not hasattr(self, "theorem") or self.theorem is None:
            return None

        prompt = (
            "### System:\n"
            "You are a Lean 4 tactic generator. Given a goal state, "
            "output exactly ONE Lean tactic that advances or solves the goal.\n"
            "Rules:\n"
            "- Output only the tactic text; no prose, quotes, or code fences.\n"
            "- Single line only; no `by` blocks.\n"
            "- Never use `sorry` or `admit`.\n"
            "### User:\n"
            "{goal_str}\n\n"
            "### Assistant:\n"
        ).format(goal_str=str(state))

        inputs = self.tokenizer(
            prompt, return_tensors="pt", truncation=True, max_length=512
        ).to(self.device)

        with torch.no_grad():
            outputs = self.model.generate(
                input_ids=inputs.input_ids,
                attention_mask=inputs.attention_mask,
                max_new_tokens=64,
                num_return_sequences=5,
                do_sample=True,
                temperature=0.7,
                top_p=0.9,
                pad_token_id=self.tokenizer.pad_token_id,
                eos_token_id=self.tokenizer.eos_token_id,
            )

        generated_texts = self.tokenizer.batch_decode(outputs, skip_special_tokens=True)

        tactics = []
        for text in generated_texts:
            tactic_part = text[len(prompt) :].strip()
            tactic_part = tactic_part.split("\n")[0].split("<;>")[0].strip()
            if tactic_part and tactic_part != "sorry":
                tactics.append(tactic_part)

        selected_tactic = random.choice(tactics)

        return selected_tactic

    def generate_whole_proof(self, theorem: Theorem) -> str:
        self.theorem = theorem

        prompt = (
            "### System:\n"
            "Given a theorem statement, "
            "output the complete proof of the theorem in Lean 4 code.\n"
            "Only output the proof, no explanation, no comments, no theorem, nothing else."
            "### User:\n"
            "{theorem_str}\n\n"
            "### Assistant:\n"
        ).format(theorem_str=str(self.theorem))

        inputs = self.tokenizer(
            prompt, return_tensors="pt", truncation=True, max_length=512
        ).to(self.device)

        with torch.no_grad():
            outputs = self.model.generate(
                input_ids=inputs.input_ids,
                attention_mask=inputs.attention_mask,
                max_new_tokens=512,
                num_return_sequences=1,
                do_sample=True,
                temperature=0.7,
                top_p=0.9,
                pad_token_id=self.tokenizer.pad_token_id,
                eos_token_id=self.tokenizer.eos_token_id,
            )

        generated_texts = self.tokenizer.batch_decode(outputs, skip_special_tokens=True)

        proof = generated_texts[0][len(prompt) :].strip().replace("<;> ", "")

        return proof



"""Cache manager of traced repos."""

import os
import shutil
import tarfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import Generator, Optional

from filelock import FileLock
from loguru import logger

from lean_dojo_v2.utils.common import execute, report_critical_failure, url_exists
from lean_dojo_v2.utils.constants import (
    CACHE_DIR,
    DISABLE_REMOTE_CACHE,
    REMOTE_CACHE_URL,
)

_CACHE_CORRPUTION_MSG = "The cache may have been corrputed!"


@dataclass(frozen=True, eq=False)
class Cache:
    """Cache manager."""

    cache_dir: Path
    lock: FileLock = field(init=False, repr=False)

    def __iter__(self) -> Generator[Path, None, None]:
        """Iterate over all traced repos in the cache."""
        yield from self.cache_dir.glob("*")

    def __post_init__(self):
        if not os.path.exists(self.cache_dir):
            self.cache_dir.mkdir(parents=True)
        lock_path = self.cache_dir.with_suffix(".lock")
        object.__setattr__(self, "lock", FileLock(lock_path))

    def get(self, rel_cache_dir: Path) -> Optional[Path]:
        """Get the cache repo at ``CACHE_DIR / rel_cache_dir`` from the cache.

        Args:
            rel_cache_dir (Path): The relative path of the stored repo in the cache.
        """
        dirname = rel_cache_dir.parent
        dirpath = self.cache_dir / dirname
        cache_path = self.cache_dir / rel_cache_dir

        with self.lock:
            if dirpath.exists():
                assert cache_path.exists()
                return cache_path

            elif not DISABLE_REMOTE_CACHE:
                url = os.path.join(REMOTE_CACHE_URL, f"{dirname}.tar.gz")
                if not url_exists(url):
                    return None
                logger.info(
                    f"Downloading the traced repo from the remote cache. Set the environment variable `DISABLE_REMOTE_CACHE` if you want to trace the repo locally."
                )
                execute(f"wget {url} -O {dirpath}.tar.gz")

                with report_critical_failure(_CACHE_CORRPUTION_MSG):
                    with tarfile.open(f"{dirpath}.tar.gz") as tar:
                        tar.extractall(self.cache_dir)
                    os.remove(f"{dirpath}.tar.gz")
                    assert (cache_path).exists()

                return cache_path

            else:
                return None

    def store(self, src: Path, rel_cache_dir: Path) -> Path:
        """Store a repo at path ``src``. Return its path in the cache.

        Args:
            src (Path): Path to the repo.
            rel_cache_dir (Path): The relative path of the stored repo in the cache.
        """
        dirpath = self.cache_dir / rel_cache_dir.parent
        cache_path = self.cache_dir / rel_cache_dir
        if not dirpath.exists():
            with self.lock:
                with report_critical_failure(_CACHE_CORRPUTION_MSG):
                    shutil.copytree(src, cache_path)
        return cache_path


cache = Cache(CACHE_DIR)
"""A global :class:`Cache` object managing LeanDojo's caching of traced repos (see :ref:`caching`).
"""



import json
import os
import random
import re
import sys
from dataclasses import dataclass, field
from typing import Any, Dict, Generator, List, Optional, Tuple

import networkx as nx
import pytorch_lightning as pl
import torch
from deepspeed.ops.adam import DeepSpeedCPUAdam, FusedAdam
from loguru import logger
from pytorch_lightning.strategies.deepspeed import DeepSpeedStrategy
from pytorch_lightning.utilities.deepspeed import (
    convert_zero_checkpoint_to_fp32_state_dict,
)
from transformers import get_cosine_schedule_with_warmup

from lean_dojo_v2.lean_dojo import Pos
from lean_dojo_v2.utils.common import zip_strict
from lean_dojo_v2.utils.constants import MARK_END_SYMBOL, MARK_START_SYMBOL

Example = Dict[str, Any]
Batch = Dict[str, Any]


@dataclass(unsafe_hash=True)
class Context:
    """Contexts are "queries" in our retrieval setup."""

    path: str
    theorem_full_name: str
    theorem_pos: Pos = field(compare=False)
    state: Optional[str] = None

    def __post_init__(self) -> None:
        assert isinstance(self.path, str)
        assert isinstance(self.theorem_full_name, str)
        assert isinstance(self.theorem_pos, Pos)
        if self.state is not None:
            if not (
                isinstance(self.state, str)
                and "⊢" in self.state
                and MARK_START_SYMBOL not in self.state
                and MARK_END_SYMBOL not in self.state
            ):
                logger.warning(f"Invalid state: {self.state}")
            assert (
                isinstance(self.state, str)
                and "⊢" in self.state
                and MARK_START_SYMBOL not in self.state
                and MARK_END_SYMBOL not in self.state
            )

    def serialize(self) -> str:
        """Serialize the context into a string for Transformers."""
        if self.state is None:
            return ""
        return self.state


def escape_regex_special_chars(text):
    return re.escape(text)


@dataclass(unsafe_hash=True)
class Premise:
    """Premises are "documents" in our retrieval setup."""

    path: str
    """The ``*.lean`` file this premise comes from.
    """

    full_name: str
    """Fully qualified name.
    """

    start: Pos = field(repr=False)
    """Start position of the premise's definition in the ``*.lean`` file.
    """

    end: Pos = field(repr=False, compare=False)
    """End position of the premise's definition in the ``*.lean`` file.
    """

    code: str = field(compare=False)
    """Raw, human-written code for defining the premise.
    """

    def __post_init__(self) -> None:
        assert isinstance(self.path, str)
        assert isinstance(self.full_name, str)
        assert (
            isinstance(self.start, Pos)
            and isinstance(self.end, Pos)
            and self.start <= self.end
        )
        assert isinstance(self.code, str) and self.code != ""

    def serialize(self) -> str:
        """Serialize the premise into a string for Transformers."""
        annot_full_name = f"{MARK_START_SYMBOL}{self.full_name}{MARK_END_SYMBOL}"
        code = self.code.replace(f"_root_.{self.full_name}", annot_full_name)
        fields = self.full_name.split(".")

        for i in range(len(fields)):
            prefix = ".".join(fields[i:])
            escaped_prefix = escape_regex_special_chars(prefix)
            pattern = f"(?<=\\s)«?{escaped_prefix}»?"
            new_code = re.sub(pattern, annot_full_name, code)
            if new_code != code:
                code = new_code
                break

        return code


class PremiseSet:
    """A set of premises indexed by their paths and full names."""

    path2premises: Dict[str, Dict[str, Premise]]

    def __init__(self) -> None:
        self.path2premises = {}

    def __iter__(self) -> Generator[Premise, None, None]:
        for _, premises in self.path2premises.items():
            for p in premises.values():
                yield p

    def add(self, p: Premise) -> None:
        if p.path in self.path2premises:
            self.path2premises[p.path][p.full_name] = p
        else:
            self.path2premises[p.path] = {p.full_name: p}

    def update(self, premises: List[Premise]) -> None:
        for p in premises:
            self.add(p)

    def __contains__(self, p: Premise) -> bool:
        return (
            p.path in self.path2premises and p.full_name in self.path2premises[p.path]
        )

    def __len__(self) -> int:
        return sum(len(premises) for premises in self.path2premises.values())


@dataclass(frozen=True)
class File:
    """A file defines 0 or multiple premises."""

    path: str
    """Path of the ``*.lean`` file.
    """

    premises: List[Premise] = field(repr=False, compare=False)
    """A list of premises defined in this file.
    """

    @classmethod
    def from_data(cls, file_data: Dict[str, Any]) -> "File":
        """Construct a :class:`File` object from ``file_data``."""
        path = file_data["path"]
        premises = []
        for p in file_data["premises"]:
            full_name = p["full_name"]
            if full_name is None:
                continue
            if "user__.n" in full_name or p["code"] == "":
                # Ignore ill-formed premises (often due to errors in ASTs).
                continue
            if full_name.startswith("[") and full_name.endswith("]"):
                # Ignore mutual definitions.
                continue
            premises.append(
                Premise(
                    path, p["full_name"], Pos(*p["start"]), Pos(*p["end"]), p["code"]
                )
            )
        return cls(path, premises)

    @property
    def is_empty(self) -> bool:
        """Check whether the file contains no premise."""
        return self.premises == []


class Corpus:
    """Our retrieval corpus is a DAG of files. Each file consists of
    premises (theorems, definitoins, etc.) that can be retrieved.
    """

    transitive_dep_graph: nx.DiGraph
    """Transitive closure of the dependency graph among files. 
    There is an edge from file X to Y iff X import Y (directly or indirectly).
    """

    all_premises: List[Premise]
    """All premises in the entire corpus.
    """

    def __init__(self, jsonl_path: str) -> None:
        """Construct a :class:`Corpus` object from a ``corpus.jsonl`` data file."""
        dep_graph = nx.DiGraph()
        self.all_premises = []

        for line in open(jsonl_path):
            file_data = json.loads(line)
            path = file_data["path"]
            assert not dep_graph.has_node(path)
            file = File.from_data(file_data)

            dep_graph.add_node(path, file=file)
            self.all_premises.extend(file.premises)

            for p in file_data["imports"]:
                assert dep_graph.has_node(p)
                dep_graph.add_edge(path, p)

        assert nx.is_directed_acyclic_graph(dep_graph)
        self.transitive_dep_graph = nx.transitive_closure_dag(dep_graph)

        self.imported_premises_cache = {}
        self.fill_cache()

    def _get_file(self, path: str) -> File:
        # for some reason, the `path` in the parameter starts with ./
        # but the paths in the corpus don't
        # so we need to remove the ./
        if path.startswith("./"):
            path = path[2:]

        if path not in self.transitive_dep_graph:
            logger.warning(f"File {path} not found in dependency graph.")
            # Return an empty file to avoid KeyError
            return File(path=path, premises=[])

        return self.transitive_dep_graph.nodes[path]["file"]

    def __len__(self) -> int:
        return len(self.all_premises)

    def __contains__(self, path: str) -> bool:
        return path in self.transitive_dep_graph

    def __getitem__(self, idx: int) -> Premise:
        return self.all_premises[idx]

    @property
    def files(self) -> List[File]:
        return [self._get_file(p) for p in self.transitive_dep_graph.nodes]

    @property
    def num_files(self) -> int:
        return len(self.files)

    def get_dependencies(self, path: str) -> List[str]:
        """Return a list of (direct and indirect) dependencies of the file ``path``."""
        return list(self.transitive_dep_graph.successors(path))

    def get_premises(self, path: str) -> List[Premise]:
        """Return a list of premises defined in the file ``path``."""
        return self._get_file(path).premises

    def num_premises(self, path: str) -> int:
        """Return the number of premises defined in the file ``path``."""
        return len(self.get_premises(path))

    def locate_premise(self, path: str, pos: Pos) -> Optional[Premise]:
        """Return a premise at position ``pos`` in file ``path``.

        Return None if no such premise can be found.
        """
        for p in self.get_premises(path):
            # accounting for . vs ./ starting in the path
            assert (p.path == path) or (p.path == path[2:])
            if p.start <= pos <= p.end:
                return p
        return None

    def fill_cache(self) -> None:
        for path in self.transitive_dep_graph.nodes:
            self._get_imported_premises(path)

    def _get_imported_premises(self, path: str) -> List[Premise]:
        """Return a list of premises imported in file ``path``. The result is cached."""
        premises = self.imported_premises_cache.get(path, None)
        if premises is not None:
            return premises

        premises = []
        for p in self.transitive_dep_graph.successors(path):
            premises.extend(self._get_file(p).premises)
        self.imported_premises_cache[path] = premises
        return premises

    def get_accessible_premises(self, path: str, pos: Pos) -> PremiseSet:
        """Return the set of premises accessible at position ``pos`` in file ``path``,
        i.e., all premises defined in the (transitively) imported files or earlier in the same file.
        """
        premises = PremiseSet()
        for p in self.get_premises(path):
            if p.end <= pos:
                premises.add(p)
        premises.update(self._get_imported_premises(path))
        return premises

    def get_accessible_premise_indexes(self, path: str, pos: Pos) -> List[int]:
        return [
            i
            for i, p in enumerate(self.all_premises)
            if (p.path == path and p.end <= pos)
            or self.transitive_dep_graph.has_edge(path, p.path)
        ]

    def get_nearest_premises(
        self,
        premise_embeddings: torch.FloatTensor,
        batch_context: List[Context],
        batch_context_emb: torch.Tensor,
        k: int,
    ) -> Tuple[List[List[Premise]], List[List[float]]]:
        """Perform a batch of nearest neighbour search."""
        similarities = batch_context_emb @ premise_embeddings.t()
        idxs_batch = similarities.argsort(dim=1, descending=True).tolist()
        results = [[] for _ in batch_context]
        scores = [[] for _ in batch_context]

        for j, (ctx, idxs) in enumerate(zip(batch_context, idxs_batch)):
            accessible_premises = self.get_accessible_premises(
                ctx.path, ctx.theorem_pos
            )
            for i in idxs:
                p = self.all_premises[i]
                if p in accessible_premises:
                    results[j].append(p)
                    scores[j].append(similarities[j, i].item())
                    if len(results[j]) >= k:
                        break
            else:
                raise ValueError

        return results, scores


@dataclass(frozen=True)
class IndexedCorpus:
    """A corpus with premise embeddings."""

    corpus: Corpus
    embeddings: torch.FloatTensor

    def __post_init__(self):
        assert self.embeddings.device == torch.device("cpu")
        assert len(self.embeddings) == len(self.corpus)


def get_all_pos_premises(annot_tac, corpus: Corpus) -> List[Premise]:
    """Return a list of all premises that are used in the tactic ``annot_tac``."""
    _, provenances = annot_tac
    all_pos_premises = set()

    for prov in provenances:
        def_path = prov["def_path"]
        p = corpus.locate_premise(def_path, Pos(*prov["def_pos"]))
        if p is not None:
            all_pos_premises.add(p)

    return list(all_pos_premises)


_SPACES_REGEX = re.compile(r"\s+", re.DOTALL)


def normalize_spaces(s: str) -> str:
    """Repalce any consecutive block of whitespace characters in ``s`` with a single whitespace."""
    return _SPACES_REGEX.sub(" ", s).strip()


def format_tactic(annot_tac: str, provenances, normalize: bool) -> str:
    """Use full names for the all <a>...</a>."""
    if normalize:
        annot_tac = normalize_spaces(annot_tac)
    if len(provenances) == 0:
        return annot_tac

    tac = ""
    marks = list(re.finditer(r"<a>(?P<ident>.+?)</a>", annot_tac))

    for i, (m, prov) in enumerate(zip_strict(marks, provenances)):
        last_end = marks[i - 1].end() if i > 0 else 0
        tac += annot_tac[last_end : m.start()] + "<a>" + prov["full_name"] + "</a>"

    tac += annot_tac[marks[-1].end() :]
    return tac


def format_state(s: str) -> str:
    m = re.match(r"\d+ goals", s)
    if m is not None:
        return s[m.end() :].strip()
    else:
        return s


def format_augmented_state(
    s: str, premises: List[Premise], max_len: int, p_drop: float
) -> str:
    """Format a state with retrieved premises and drop some of them with probability ``p_drop``."""
    s = format_state(s)

    aug_s = ""
    length = 0
    max_premises_len = max_len - len(bytes(s.encode("utf-8")))

    for p in premises:
        if random.random() < p_drop:
            continue
        p_str = f"{p.serialize()}\n\n"
        l = len(bytes(p_str.encode("utf-8")))
        if length + l > max_premises_len:
            continue
        length += l
        aug_s = p_str + aug_s

    aug_s += s
    return aug_s


def get_optimizers(
    parameters, trainer: pl.Trainer, lr: float, warmup_steps: int
) -> Dict[str, Any]:
    """Return an AdamW optimizer with cosine warmup learning rate schedule."""
    strategy = trainer.strategy

    if isinstance(strategy, DeepSpeedStrategy):
        if "offload_optimizer" in strategy.config["zero_optimization"]:
            logger.info("Optimizing with DeepSpeedCPUAdam")
            optimizer = DeepSpeedCPUAdam(parameters, lr=lr, adamw_mode=True)
        else:
            logger.info("Optimizing with FusedAdam")
            optimizer = FusedAdam(parameters, lr=lr, adam_w_mode=True)
    else:
        logger.info("Optimizing with AdamW")
        optimizer = torch.optim.AdamW(parameters, lr=lr)

    if trainer.max_steps != -1:
        max_steps = trainer.max_steps
    else:
        assert trainer.max_epochs is not None
        max_steps = (
            trainer.max_epochs
            * len(trainer.datamodule.train_dataloader())
            // trainer.accumulate_grad_batches
        )

    scheduler = get_cosine_schedule_with_warmup(
        optimizer,
        num_warmup_steps=warmup_steps,
        num_training_steps=max_steps,
    )

    return {
        "optimizer": optimizer,
        "lr_scheduler": {
            "scheduler": scheduler,
            "interval": "step",
        },
    }



#!/usr/bin/env python3
"""
Fine-tunes a LeanProgress regression head that predicts steps remaining.
"""

from __future__ import annotations

import argparse
from pathlib import Path

import numpy as np
from datasets import load_dataset
from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
    Trainer,
    TrainingArguments,
)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Train a LeanProgress regression model."
    )
    parser.add_argument(
        "--dataset",
        type=Path,
        required=True,
        help="JSONL with goal/prefix/tactic/steps_remaining fields.",
    )
    parser.add_argument(
        "--output-dir",
        type=Path,
        required=True,
        help="Directory to save the fine-tuned model.",
    )
    parser.add_argument(
        "--model-name",
        type=str,
        default="bert-base-uncased",
        help="Base Hugging Face model to fine-tune.",
    )
    parser.add_argument("--max-length", type=int, default=512)
    parser.add_argument("--batch-size", type=int, default=8)
    parser.add_argument("--epochs", type=float, default=3.0)
    parser.add_argument("--learning-rate", type=float, default=1e-5)
    parser.add_argument("--eval-ratio", type=float, default=0.2)
    parser.add_argument("--seed", type=int, default=42)
    return parser.parse_args()


def build_text(goal: str, prefix: str, tactic: str) -> str:
    prefix = prefix or ""
    return (
        "Goal:\n" + goal + "\n\nPrefix:\n" + prefix + "\n\nCandidate tactic:\n" + tactic
    )


def tokenize_batch(tokenizer, max_length):
    def _tokenize(batch):
        texts = [
            build_text(goal, prefix, tactic)
            for goal, prefix, tactic in zip(
                batch["goal"],
                batch.get("prefix", [""] * len(batch["goal"])),
                batch["tactic"],
            )
        ]
        encoded = tokenizer(
            texts,
            padding="max_length",
            truncation=True,
            max_length=max_length,
        )
        encoded["labels"] = [float(x) for x in batch["steps_remaining"]]
        return encoded

    return _tokenize


def compute_metrics(eval_pred):
    predictions, labels = eval_pred
    preds = predictions.squeeze()
    mse = ((preds - labels) ** 2).mean().item()
    mae = (np.abs(preds - labels)).mean().item()
    return {"mse": mse, "mae": mae}


def main() -> None:
    args = parse_args()
    raw_dataset = load_dataset(
        "json",
        data_files={"train": str(args.dataset)},
    )["train"]
    dataset = raw_dataset.train_test_split(
        test_size=args.eval_ratio,
        seed=args.seed,
        stratify_by_column=None,
    )

    tokenizer = AutoTokenizer.from_pretrained(args.model_name)
    if tokenizer.pad_token is None:
        tokenizer.pad_token = tokenizer.eos_token or tokenizer.cls_token

    tokenized = dataset.map(
        tokenize_batch(tokenizer, args.max_length),
        batched=True,
        remove_columns=dataset["train"].column_names,
    )

    model = AutoModelForSequenceClassification.from_pretrained(
        args.model_name,
        num_labels=1,
        problem_type="regression",
    )

    training_args = TrainingArguments(
        output_dir=str(args.output_dir),
        learning_rate=args.learning_rate,
        per_device_train_batch_size=args.batch_size,
        per_device_eval_batch_size=args.batch_size,
        num_train_epochs=args.epochs,
        weight_decay=0.01,
        eval_strategy="epoch",
        save_strategy="epoch",
        load_best_model_at_end=True,
        metric_for_best_model="mse",
        logging_steps=50,
        report_to="none",
        seed=args.seed,
    )

    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=tokenized["train"],
        eval_dataset=tokenized["test"],
        tokenizer=tokenizer,
        compute_metrics=compute_metrics,
    )
    trainer.train()
    trainer.save_model(args.output_dir)
    tokenizer.save_pretrained(args.output_dir)
    print(
        f"LeanProgress model saved to {args.output_dir.resolve()}. "
        "Set LEANPROGRESS_MODEL to this path to enable step predictions."
    )


if __name__ == "__main__":
    main()



"""Prover package for Lean theorem proving agents."""

from .base_prover import BaseProver
from .external_prover import ExternalProver
from .hf_prover import HFProver
from .retrieval_prover import RetrievalProver

__all__ = ["BaseProver", "ExternalProver", "HFProver", "RetrievalProver"]



"""This module provides the main interfaces for tracing Lean repos, i.e., extracting data from them.
To estimate the time for tracing a repo, a good rule of thumb is 1.5x the time for compiling the repo using :code:`leanpkg build`.
A repo has to be traced only once, and the traced repo will be stored in a cache for fast access in the future.
"""

import itertools
import os
import re
import shutil
from contextlib import contextmanager
from multiprocessing import Process
from pathlib import Path
from subprocess import CalledProcessError
from time import monotonic, sleep
from typing import Generator, List, Optional, Union

from loguru import logger
from tqdm import tqdm

from lean_dojo_v2.utils.common import execute
from lean_dojo_v2.utils.constants import NUM_PROCS
from lean_dojo_v2.utils.filesystem import working_directory

from .cache import cache
from .lean import LeanGitRepo
from .traced_data import TracedRepo

LEAN4_DATA_EXTRACTOR_PATH = Path(__file__).with_name("ExtractData.lean")

_PROGRESSBAR_UPDATE_INTERNAL = 5


def _modify_dependency_files(packages_path: Path) -> None:
    """Modify dependency files to replace 'import all' with 'public import all'."""
    logger.debug(
        "Modifying dependency files to replace 'import all' with 'public import all'"
    )
    for lean_file in packages_path.rglob("*.lean"):
        with open(lean_file, "r", encoding="utf-8") as f:
            content = f.read()

        modified_content = re.sub(
            r"^(\s*)import all", r"\1public import all", content, flags=re.MULTILINE
        )

        if modified_content != content:
            with open(lean_file, "w", encoding="utf-8") as f:
                f.write(modified_content)


def _monitor(paths: List[Path], num_total: int) -> None:
    with tqdm(total=num_total) as pbar:
        while True:
            time_start = monotonic()
            try:
                num_done = len(
                    list(
                        itertools.chain.from_iterable(
                            p.glob(f"**/*.ast.json") for p in paths
                        )
                    )
                )
            except Exception:
                continue
            time_elapsed = monotonic() - time_start
            if time_elapsed < _PROGRESSBAR_UPDATE_INTERNAL:
                sleep(_PROGRESSBAR_UPDATE_INTERNAL - time_elapsed)
            pbar.update(num_done - pbar.n)
            if num_done >= num_total:
                break
    print("")


@contextmanager
def launch_progressbar(paths: List[Path]) -> Generator[None, None, None]:
    """Launch an async progressbar to monitor the progress of tracing the repo."""
    paths = [Path(p) for p in paths]
    olean_files = list(
        itertools.chain.from_iterable(p.glob("**/*.olean") for p in paths)
    )
    num_total = len(olean_files)
    p = Process(target=_monitor, args=(paths, num_total), daemon=True)
    p.start()
    yield
    p.kill()


def get_lean_version() -> str:
    """Get the version of Lean."""
    output = execute("lean --version", capture_output=True)[0].strip()
    m = re.match(r"Lean \(version (?P<version>\S+?),", output)
    return m["version"]  # type: ignore


def is_new_version(v: str) -> bool:
    """Check if ``v`` is at least `4.3.0-rc2`."""
    major, minor, patch = [int(_) for _ in v.split("-")[0].split(".")]
    if major < 4 or (major == 4 and minor < 3):
        return False
    if (
        major > 4
        or (major == 4 and minor > 3)
        or (major == 4 and minor == 3 and patch > 0)
    ):
        return True
    assert major == 4 and minor == 3 and patch == 0
    if "4.3.0-rc" in v:
        rc = int(v.split("-")[1][2:])
        return rc >= 2
    else:
        return True


def check_files(packages_path: Path, no_deps: bool) -> None:
    """Check if all :file:`*.lean` files have been processed to produce :file:`*.ast.json` and :file:`*.dep_paths` files."""
    cwd = Path.cwd()
    packages_path = cwd / packages_path
    jsons = {
        p.with_suffix("").with_suffix("")
        for p in cwd.glob("**/build/ir/**/*.ast.json")
        if not no_deps or not p.is_relative_to(packages_path)
    }
    deps = {
        p.with_suffix("")
        for p in cwd.glob("**/build/ir/**/*.dep_paths")
        if not no_deps or not p.is_relative_to(packages_path)
    }
    oleans = {
        Path(str(p.with_suffix("")).replace("/build/lib/lean/", "/build/ir/"))
        for p in cwd.glob("**/build/lib/lean/**/*.olean")
        if not no_deps or not p.is_relative_to(packages_path)
    }
    assert len(jsons) <= len(oleans) and len(deps) <= len(oleans)
    missing_jsons = {p.with_suffix(".ast.json") for p in oleans - jsons}
    missing_deps = {p.with_suffix(".dep_paths") for p in oleans - deps}
    if len(missing_jsons) > 0 or len(missing_deps) > 0:
        for p in missing_jsons.union(missing_deps):
            logger.warning(f"Missing {p}")


def _trace(repo: LeanGitRepo, build_deps: bool) -> None:
    assert (
        repo.exists()
    ), f"The {repo} does not exist. Please check the URL `{repo.commit_url}`."

    # Trace `repo` in the current working directory.
    assert not repo.is_lean4, "Cannot trace Lean 4 itself."
    if not os.path.exists(repo.name):
        repo.clone_and_checkout()
    logger.debug(f"Tracing {repo}")

    with working_directory(repo.name):
        # Build the repo using lake.
        if not build_deps:
            try:
                execute("lake exe cache get")
            except CalledProcessError:
                pass
        execute("lake build")

        # Copy the Lean 4 stdlib into the path of packages.
        lean_prefix = execute(f"lean --print-prefix", capture_output=True)[0].strip()
        if is_new_version(get_lean_version()):
            packages_path = Path(".lake/packages")
            build_path = Path(".lake/build")
        else:
            packages_path = Path("lake-packages")
            build_path = Path("build")

        shutil.copytree(
            Path(lean_prefix), str(packages_path / "lean4"), dirs_exist_ok=True
        )

        # Modify dependency files to replace 'import all' with 'public import all'
        if build_deps:
            _modify_dependency_files(packages_path)

        # Run ExtractData.lean to extract ASTs, tactic states, and premise information.
        shutil.copyfile(LEAN4_DATA_EXTRACTOR_PATH, LEAN4_DATA_EXTRACTOR_PATH.name)

        dirs_to_monitor = [build_path]
        if build_deps:
            dirs_to_monitor.append(packages_path)
        with launch_progressbar(dirs_to_monitor):
            cmd = f"lake env lean --threads {NUM_PROCS} --run ExtractData.lean"
            if not build_deps:
                cmd += " noDeps"
            execute(cmd)

        check_files(packages_path, not build_deps)
        os.remove(LEAN4_DATA_EXTRACTOR_PATH.name)


def is_available_in_cache(repo: LeanGitRepo) -> bool:
    """Check if ``repo`` has a traced repo available in the cache (including the remote cache)."""
    rel_cache_dir = repo.get_cache_dirname() / repo.name
    return cache.get(rel_cache_dir) is not None


def get_traced_repo_path(repo: LeanGitRepo, build_deps: bool = True) -> Path:
    """Return the path of a traced repo in the cache.

    The function will trace a repo if it is not available in the cache. See :ref:`caching` for details.

    Args:
        repo (LeanGitRepo): The Lean repo to trace.
        build_deps (bool): Whether to build the dependencies of ``repo``. Defaults to True.

    Returns:
        Path: The path of the traced repo in the cache, e.g. :file:`/home/kaiyu/.cache/lean_dojo/leanprover-community-mathlib-2196ab363eb097c008d4497125e0dde23fb36db2`
    """
    rel_cache_dir = repo.get_cache_dirname() / (
        repo.name + ("_d" if build_deps else "")
    )

    path = cache.get(rel_cache_dir)

    if path is None and not build_deps:
        path = cache.get(rel_cache_dir.parent / (repo.name + "_d"))

    if path is None:
        logger.info(f"Tracing {repo}")
        with working_directory() as tmp_dir:
            logger.debug(f"Working in the temporary directory {tmp_dir}")
            _trace(repo, build_deps)
            src_dir = tmp_dir / repo.name
            traced_repo = TracedRepo.from_traced_files(src_dir, build_deps)
            traced_repo.save_to_disk()
            path = cache.store(src_dir, rel_cache_dir)
    else:
        logger.debug("The traced repo is available in the cache.")
    return path


def trace(
    repo: LeanGitRepo,
    dst_dir: Optional[Union[str, Path]] = None,
    build_deps: bool = True,
) -> TracedRepo:
    """Trace a repo (and its dependencies), saving the results to ``dst_dir``.

    The function only traces the repo when it's not available in the cache. Otherwise,
    it directly copies the traced repo from the cache to ``dst_dir``. See :ref:`caching` for details.

    Args:
        repo (LeanGitRepo): The Lean repo to trace.
        dst_dir (Union[str, Path]): The directory for saving the traced repo. If None, the traced repo is only saved in the cahe.
        build_deps (bool): Whether to build the dependencies of ``repo``. Defaults to True.

    Returns:
        TracedRepo: A :class:`TracedRepo` object corresponding to the files at ``dst_dir``.
    """
    if dst_dir is not None:
        dst_dir = Path(dst_dir)
        assert (
            not dst_dir.exists()
        ), f"The destination directory {dst_dir} already exists."

    cached_path = get_traced_repo_path(repo, build_deps)
    logger.info(f"Loading the traced repo from {cached_path}")
    traced_repo = TracedRepo.load_from_disk(cached_path, build_deps)
    traced_repo.check_sanity()

    if dst_dir is not None:
        dst_dir.mkdir(parents=True)
        shutil.copytree(cached_path, dst_dir / cached_path.name, dirs_exist_ok=True)

    return traced_repo



import json
from types import SimpleNamespace

from lean_dojo_v2.lean_progress import create_sample_dataset, train_steps_model


def test_create_sample_dataset_writes_expected_jsonl(tmp_path, monkeypatch):
    output_path = tmp_path / "sample.jsonl"
    monkeypatch.setattr(
        create_sample_dataset, "parse_args", lambda: SimpleNamespace(output=output_path)
    )
    create_sample_dataset.main()

    with output_path.open() as f:
        rows = [json.loads(line) for line in f]
    assert rows == create_sample_dataset.SAMPLE_DATA


def test_build_text_structure_contains_sections():
    payload = {
        "goal": "gcd n n = n",
        "prefix": "intro n",
        "tactic": "simpa",
    }
    text = train_steps_model.build_text(
        payload["goal"], payload["prefix"], payload["tactic"]
    )
    assert "Goal:\n" in text
    assert "Prefix:\n" in text
    assert "Candidate tactic:\n" in text
    assert payload["goal"] in text
    assert payload["prefix"] in text
    assert payload["tactic"] in text


class DummyTokenizer:
    def __call__(self, texts, padding, truncation, max_length):
        self.last_texts = texts
        assert padding == "max_length"
        assert truncation is True
        assert isinstance(max_length, int)
        return {"input_ids": list(range(len(texts)))}


def test_tokenize_examples_handles_missing_prefix():
    tokenizer = DummyTokenizer()
    tokenize_fn = train_steps_model.tokenize_batch(tokenizer, max_length=32)
    batch = {
        "goal": ["goal 1", "goal 2"],
        # leave out prefix to trigger default handling
        "tactic": ["tac1", "tac2"],
        "steps_remaining": [3, 0],
    }

    encoded = tokenize_fn(batch)
    assert encoded["labels"] == batch["steps_remaining"]
    assert len(tokenizer.last_texts) == 2
    assert "goal 1" in tokenizer.last_texts[0]
    assert "tac1" in tokenizer.last_texts[0]



import os
from abc import ABC, abstractmethod
from pathlib import Path

from loguru import logger
from pantograph import Server

from lean_dojo_v2.database import DynamicDatabase
from lean_dojo_v2.database.models import Repository
from lean_dojo_v2.lean_dojo.data_extraction.lean import LeanGitRepo
from lean_dojo_v2.lean_dojo.data_extraction.trace import get_traced_repo_path
from lean_dojo_v2.utils.constants import DATA_DIR, RAID_DIR


class BaseAgent(ABC):
    """Abstract base class for theorem proving agents.

    This class defines the common interface and functionality for different types
    of theorem proving agents, such as HFAgent and LeanAgent.
    """

    def __init__(self, database_path: str = "dynamic_database.json"):
        self.database = DynamicDatabase(json_path=database_path)
        self.data_path = Path(os.path.join(RAID_DIR, DATA_DIR, "merged"))
        self.repos = []

    @abstractmethod
    def _get_build_deps(self) -> bool:
        """Get whether to build dependencies. Must be implemented by subclasses."""
        pass

    def trace_repository(self, url: str, commit: str, build_deps: bool) -> Repository:
        """Trace a repository and return a Repository object."""
        return self.database.trace_repository(url, commit, build_deps)

    def add_repository(self, repo: Repository):
        """Add a repository to the database."""
        self.database.add_repository(repo)
        self.repos.append(LeanGitRepo(repo.url, repo.commit))

    def train(self):
        """Train the model on the repository.

        Raises:
            ValueError: If no repository is loaded
        """
        sorted_repos = self.database.sort_repositories_by_difficulty()

        if len(sorted_repos) == 0:
            raise ValueError(
                "No repository loaded. Call setup_github_repository() or setup_local_repository() first."
            )

        self.trainer.train(
            repos=sorted_repos, database=self.database, data_path=self.data_path
        )

    def evaluate(self):
        """Evaluate the trained model."""
        self.trainer.evaluate()

    def setup_github_repository(self, url: str, commit: str):
        """Set up a GitHub repository for processing."""
        traced_repo = self.trace_repository(
            url, commit, build_deps=self._get_build_deps()
        )
        if traced_repo:
            self.add_repository(traced_repo)
        else:
            raise ValueError(f"Failed to setup github repository {url}")

    def setup_local_repository(self, path: str):
        """Set up a local repository for processing."""
        repo = LeanGitRepo.from_path(path)
        traced_repo = self.trace_repository(
            repo.url, repo.commit, build_deps=self._get_build_deps()
        )
        if traced_repo:
            self.add_repository(traced_repo)
        else:
            raise ValueError(f"Failed to setup local repository {path}")

    def initialize_prover(self):
        """Initialize the theorem prover.

        Returns:
            List of sorry theorems to prove
        """
        sorry_theorems = []
        for repo in self.repos:
            repository = self.database.get_repository(repo.url, repo.commit)
            for theorem in repository.sorry_theorems_unproved:
                sorry_theorems.append((theorem, repo))

        self._setup_prover()

        return sorry_theorems

    @abstractmethod
    def _setup_prover(self):
        """Set up the prover agent. Must be implemented by subclasses."""
        pass

    def prove(self, whole_proof: bool = False):
        """Prove sorry theorems."""
        sorry_theorems = self.initialize_prover()

        if not sorry_theorems:
            print("No sorry theorems found to prove.")
            return

        print(f"Found {len(sorry_theorems)} sorry theorems to prove")
        for theorem, repo in sorry_theorems:
            self.prove_theorem(theorem, repo, whole_proof)

    def prove_theorem(self, theorem, repo, whole_proof: bool = False):
        """Processes a single theorem."""
        if whole_proof:
            proof = self.prover.generate_whole_proof(theorem)
            print(proof)
            return

        traced_repo_path = get_traced_repo_path(repo, build_deps=self._get_build_deps())

        server = Server(
            imports=["Init", str(theorem.file_path).replace(".lean", "")],
            project_path=traced_repo_path,
        )

        print(f"Proving {theorem.full_name}")
        result, used_tactics = self.prover.search(
            server=server, theorem=theorem, verbose=False
        )
        print(result)
        if result.success:
            for tactic in used_tactics:
                print(tactic)
        else:
            logger.info(f"No proof found for {theorem.full_name}")

        return result.success
